// VinciaFunctions.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2012 P.Z.Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's Functions class.
  
#ifndef VINCIAFUNCTIONS_H
#define VINCIAFUNCTIONS_H

#include "Basics.h"                   // Vec4
#include "Event.h"                    // Particle
#include "VinciaCommon.h"             // Common
#include "VinciaTools.h"              // VinClu

using namespace Pythia8;

namespace Vincia {

  class Functions {

  public: 

    // Constructor
    Functions() {};

    // Initialize Pointers
    void initPtr(VinClu* vinCluPtrIn) {
      vinCluPtr      = vinCluPtrIn; 
      isInitPtr=true;
    }

    // *********************************************************************
    // Auxiliary Hardcoded Matrix Elements and Antenna Functions
    // *********************************************************************
    
    // Eikonal (without color or coupling factors)
    double Eikonal(Vec4 a, Vec4 r, Vec4 b) {
      double sar = m2(a+r);
      double srb = m2(r+b);
      double sab = m2(a+b);
      return 2*sab/(sar*srb);
    }
    double Eikonal(Particle a, Particle r, Particle b) {
      return Eikonal(a.p(), r.p(), b.p());
    }
    double Eikonal(Particle* a, Particle* r, Particle* b) {
      return Eikonal(a->p(), r->p(), b->p());
    }

    // GGG functions 
    
    // 3-parton tree-level GGG antennae
    double A30(double sar,double srb,double s);
    double A30(double sar,double srb,double sab,double s);
    double sA30(double yar,double yrb);
    double A30(Vec4,Vec4,Vec4);
    double A30(Particle a,Particle r,Particle b) {return A30(a.p(), r.p(), b.p());}
    double A30(Particle* a,Particle* r,Particle* b) {
      return A30(a->p(), r->p(), b->p());}
    double A30(vector<Vec4>&,int,int,int);
    double D30(double sar,double srb,double sab,double s);
    double E30(double sar,double srb,double sab,double s);
    double F30(double sar,double srb, double sab, double s);
    double G30(double sar,double srb,double sab,double s);
    double Q30(double sar,double srb,double s);
    // 3-parton tree-level GGG antennae, singular terms only
    double A30sng(double sar,double srb,double sab,double s);
    double D30sng(double sar,double srb,double sab,double s);
    double E30sng(double sar,double srb,double sab,double s);
    double F30sng(double sar,double srb, double sab, double s);
    double G30sng(double sar,double srb,double sab,double s);
    // 3-parton tree-level Skands antennae
    double E30pzs(double sar,double srb,double sab,double s);
    // 3-parton tree-level default ("Med") VINCIA antennae
    double E30med(double sar,double srb, double sab, double s);
    // 4-parton tree-level GGG antennae
    double A40(vector<Vec4>& pn, int a, int b, int c, int d);
    double A40aux(vector<Vec4>& pn, int a,int b,int c,int d);
    double A40NLC(vector<Vec4>& pn, int a, int b, int c, int d);
    double A40auxNLC(vector<Vec4>& pn, int a,int b,int c,int d);
    double B40(vector<Vec4>& pn, int a, int b, int c, int d);
    double B40aux(vector<Vec4>& pn, int a, int b, int c, int d);
    double C40(vector<Vec4>& pn, int a, int b, int c, int d);
    double C40aux(vector<Vec4>& pn, int a, int b, int c, int d);
    double D40(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40a(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40b(vector<Vec4>& pn, int a, int b, int c, int d);
    double F40(vector<Vec4>& pn, int a, int b, int c, int d);
    double G40(vector<Vec4>& pn, int a, int b, int c, int d);
    double G40NLC(vector<Vec4>& pn, int a, int b, int c, int d);
    double H40(vector<Vec4>& pn, int a, int b, int c, int d);
    // 3-parton 1-loop GGG antennae (PS-subtracted)
    double A31fin(double s12,double s23,double s13,double s123);
    double A31finNLC(double s12,double s23,double s13,double s123);
    double A31auxR(double y, double z);
    
    // *********************************************************************
    // Ariadne functions
    double A30ar(double sar, double srb, double s);
    double D30ar(double sar, double srb, double s);
    double E30ar(double sar, double srb, double s);
    double F30ar(double sar, double srb, double s);
    
    // *********************************************************************
    // Functions imported from the GGG form code, for counter-checks
    // A40LC checked, rel diffs < 1e-15.
    // A40NLC sum (= complete ~A40) checked, rel diffs < 1e-15.
    // B40 checked against form code and against Ellis,Ross,Terrano, d < 1e-13. 
    // C40 checked against form code
    // D40 (above) imported directly from form code
    // E40 checked against form code, rel diffs < 1e-15.
    // H40 checked against form code, rel diffs < 1e-14.
    
    double A40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double A40NLCfrm(vector<Vec4>& pn, int a, int b, int c, int d);
    double B40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double C40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    double E40NLCfrm(vector<Vec4>& pn, int a, int b, int c, int d);
    double H40frm(vector<Vec4>& pn, int a, int b, int c, int d);
    
    // *********************************************************************
    // Functions copied from Ellis, Ross, and Terrano
    
    double B40ert(vector<Vec4>& pn, int a, int b, int c, int d);
    
    // *********************************************************************
    // Kosower's sector functions
    
    double Asec30gks(double sar,double srb,double s);
    double Dsec30gks(double sar,double srb,double s);
    double Esec30gks(double sar,double srb,double s);
    double Fsec30gks(double sar,double srb,double s);
    double Gsec30gks(double sar,double srb,double s);
    
    // *********************************************************************
    // Skands' experimental sector functions
    
    double Asec30ska(double sar,double srb,double s);
    double Dsec30ska(double sar,double srb,double s);
    double Esec30ska(double sar,double srb,double s);
    double Fsec30ska(double sar,double srb,double s);
    double Gsec30ska(double sar,double srb,double s);
   
  private:

    // Pointer to vinClu (used by some sub-antennae)
    VinClu*            vinCluPtr;
    bool isInitPtr;

  };

}

#endif
