// VinciaMatching.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2012 P.Z.Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's Matching class.
  
#ifndef VINCIAMATCHING_H
#define VINCIAMATCHING_H

#include "Basics.h"                   // Vec4
#include "VinciaAntenna.h"            // Antenna
#include "VinciaCommon.h"             // Common
#include "VinciaEvolutionVariables.h" // EvolutionVariable
#include "VinciaTools.h"              // VinClu
#include "VinciaMGInterface.h"        // MGInterface

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  //
  // class Matching : for computing matching coefficients
  // 
  // Is structured like many other VINCIA classes, with a constructor that only
  // initializes a few init flags to false, so that we may know if the class
  // has not been properly initialized. There are three initialization routines 
  // which must be called before a Matching object can really be used: first 
  // initAntPtr() and initPtr() and then init(). The former two set the 
  // addresses to the various other objects that the matching class needs 
  // to function. In particular, the call to initAntPtr tells the matching class
  // where to look for the AntennaSet objects that will be used in the current 
  // run. The latter, init(), initializes the matching class itself, i.e., sets
  // its internal variables, etc. 
  // The class contains the following main usable functions:
  //    isMatched() : to tell whether there is any matching available for 
  //                  a given final state. 
  //    getPME()    : to give the matching coefficient for a given final
  //                  state, using the same shower functions as the rest
  //                  of the code (via the AntennaSet pointers). 
  //    getP2to4()  : FOR DEVELOPMENT ONLY.
  //
  // More detailed descriptions are given below, with each function.
  //
  //***********************************************************************

  class Matching {
    
  public:
    
    // Constructor.
    Matching() {isInit=false; isInitPtr=false;};

    // Destructor.
    virtual ~Matching() {};

    // Initialize Pointers (must be done before init)
    void initPtr(EvolutionVariable* RPtrIn, MGInterface* mgInterfacePtrIn, 
		 Settings* settingsPtrIn, VinClu* vinCluPtrIn, 
		 VinciaCommon* vinComPtrIn) {
      RPtr           = RPtrIn;
      mgInterfacePtr = mgInterfacePtrIn; 
      settingsPtr    = settingsPtrIn; 
      vinCluPtr      = vinCluPtrIn; 
      vinComPtr      = vinComPtrIn;
      isInitPtr=true;
    }

    // Initialize Pointers to Antenna Sets (must be done before init)
    void initAntPtr(AntennaSet* antV0, AntennaSet* antV1=0, 
		    AntennaSet* antV2=0, AntennaSet* antV3=0) {
      antennaSets.resize(0);
      antennaSets.push_back(antV0);
      if (antV1 != 0) antennaSets.push_back(antV1);
      if (antV2 != 0) antennaSets.push_back(antV2);
      if (antV3 != 0) antennaSets.push_back(antV3);
    }

    // Initialize
    bool init(int nVariationsIn=1);

    // Main functions to return whether and how far a specific decaying particle
    // is matched

    // *********************************************************************
    // isMatched() : return true if any matching exists at this order
    // Ideally, should replace nEmit, nSplit by nGluon, nQuark.
    // Also need a version that applies to a 2->n process, not just res dec.
    // *********************************************************************

    bool isMatched(int idRes, int nEmit, int nSplit, bool tryMG=false);

    // *********************************************************************
    // getPME() : Return a matching coefficient
    // INPUT: 
    //      nIn : 1 (for decay) or 2 (for 2->2 process)
    //      id  : vector containing PDG ID codes, 
    //           first incoming parton(s), then outgoing ones
    //      p   : same as is, for parton 4-vectors
    //   Note   : the parton list is assumed to be color-ordered
    // OUTPUT: 
    //      PME : vector whose [0] component is the matching coefficient
    //            for the current shower settings, with the additional ones
    //            corresponding to uncertainty variations (if switched on).
    // *********************************************************************

    bool getPME(int nIn, vector<int> id, vector<Vec4> p, vector<double>& PME);

    // *********************************************************************
    // getP2to4() : Return a 2->4 matching coefficient (under development)
    // *********************************************************************

    bool getP2to4(vector<int> id, vector<Vec4> p, vector<double>& PME);

    // *********************************************************************
    // getAntTree() : return |M_n|^2/|M_Born|^2 for these daughters
    // Not used for the time being so commented out
    // *********************************************************************

    // double getAntTree(int idRes, vector<Particle> daughters);

    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;}

  private:

    // Is initialized?    
    bool isInit, isInitPtr;

    // Color Depth to use in matching (0:Leading, 1:Full)
    bool fullColor;

    // Is MG Matching switched on?
    bool isMGOn;

    // Uncertainty Bands
    bool uncertaintyBands;
    int nVariations;

    // Pointers to PYTHIA and VINCIA objects
    EvolutionVariable* RPtr;
    MGInterface*       mgInterfacePtr;
    Settings*          settingsPtr;
    VinClu*            vinCluPtr;
    VinciaCommon*      vinComPtr;

    // Vector of pointers to antenna sets (0 is physical, >= 1 are variations)
    vector<AntennaSet*> antennaSets;    

    // Ordering Mode
    int evolutionMode;

    // Shower NLC 
    int showerNLC;

    // Sector ordering
    bool sectorOrdering;

    // Verbose level
    int verbose;

  };

}

#endif
