// VinciaRoot.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2012 P.Z.Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Vincia ROOT utilities

#ifndef VINCIAROOT_H
#define VINCIAROOT_H

// Include specific ROOT headers 

// Main ROOT header
#include "TROOT.h"
// 1D and 2D histograms
#include "TH1.h"
#include "TH2.h"
#include "TGraphAsymmErrors.h"
#include "TPaveStats.h"
#include "TLatex.h"
#include "TLegend.h"
#include "TLegendEntry.h"
#include "TLine.h"
// Widget windows
#include "TCanvas.h"
#include "TApplication.h"
// To read/write files
#include "TDirectory.h"
#include "TFile.h"

// Also include the VRFile formats
#include "VinciaRootFile.h"

// Put all the rest in Vincia namespace

namespace Vincia {
  
  // ------------------------------------------------------------------------
  // class VRsettings
  // Contains global settings for the VinciaRoot interface

  class VRsettings {

  public:

    // Constructor
    VRsettings() : isInitApp(false), logY(false), 
      normalize(true), showRatioPad(true), showRelUncPads(true), 
      verbose(1), optStat(1101), sigmaStat(2.), normFac(1.),
      MCnames(" "), signature("V I N C I A R O O T") { legend.resize(0); };
    
    // Data members
    bool isInitApp;
    bool logY, normalize;
    bool showRatioPad, showRelUncPads;
    int verbose, optStat;
    double sigmaStat, normFac;
    string MCnames, signature;
    vector<string> legend;

  };
  
  // ------------------------------------------------------------------------
  // class VRgraph
  // Base class for VinciaRoot graphs
  
  class VRgraph : public TGraphAsymmErrors {

  public:

    // Constructors
    VRgraph() { 
      hasOuterSav = false; 
    }
    VRgraph(TH1D* initHis) { 
      hisPtrSav.push_back(initHis); 
      hasOuterSav = false;
    }
    VRgraph(vector<TH1D*> initHis) { 
      hisPtrSav = initHis;
      hasOuterSav = (hisPtrSav.size() >= 2) ? true : false;
    }
    
    // Outer band
    bool hasOuter() {return hasOuterSav;}
    int nOuter() { return hisPtrSav.size() -1; }
    TGraphAsymmErrors* getOuter() {return &outer;}
    virtual void DrawOuter(string drawOptions="") {
      if (hasOuter()) outer.Draw(drawOptions.c_str());
    }

    // Update graph
    virtual bool update(bool normalizeIn=true, double normFacIn=1.0, 
      double sigmaStat=2.0, double scatterX = 0.0) { 

      // Check histogram not empty
      if (hisPtrSav[0]->GetEntries() <= 0) return false;

      // Update
      sumWeights   = 0.0;
      // Include overflow bins in normalization
      int nBin = hisPtrSav[0]->GetNbinsX();
      int firstBin = 0;
      int nPoints  =  -1;
      for (int iBin = 0; iBin <= nBin; ++iBin) {
	double x    = hisPtrSav[0]->GetBinCenter(iBin);
	double exLo = x - hisPtrSav[0]->GetBinLowEdge(iBin);
	double exHi = hisPtrSav[0]->GetBinWidth(iBin) - exLo;
	// Apply scatter in x (so as not to plot in exactly same point)
	if (scatterX > 0.) x += scatterX * exHi;
	else x += scatterX * exLo;
	double y    = hisPtrSav[0]->GetBinContent(iBin);
	double ey   = sigmaStat * hisPtrSav[0]->GetBinError(iBin); 
	if (iBin == firstBin && y == 0.0) {
	  ++firstBin;
	  continue;
	}
	else if (y == 0. && ey == 0.) continue;
	// Count up graph point number from first non-overflow bin
	else if (iBin >= 1) {
	  ++nPoints;
	}
	// Add uncertainty variations in quadrature with statistical
	double eyOutHi = 0.;
	double eyOutLo = 0.;
	if (hasOuter()) {
	  for (int iHis = 1; iHis < int(hisPtrSav.size()); ++iHis) {
	    // Check if same bins, else give error and switch off uncertainties
	    if (abs(hisPtrSav[iHis]->GetBinCenter(iBin) - x) > TINY || 
		abs(x - hisPtrSav[iHis]->GetBinLowEdge(iBin) - exLo) > TINY ||
		abs(hisPtrSav[iHis]->GetBinWidth(iBin) - (exLo+exHi)) > TINY) {
	      cout<<"(VRgraph): incompatible bins for uncertainty histogram "<<iHis
		  <<" for "<<hisPtrSav[0]->GetName()<<endl;
	      hasOuterSav = false;
	      break;
	    }
	    // Compute error bracket
	    double yBand = hisPtrSav[iHis]->GetBinContent(iBin);
	    if (yBand > y) eyOutHi = max(eyOutHi,yBand-y);
	    else eyOutLo = max(eyOutLo,y-yBand);	    
	  }	  
	  eyOutHi = sqrt(pow(ey,2)+pow(eyOutHi,2));
	  eyOutLo = sqrt(pow(ey,2)+pow(eyOutLo,2));
	  if (iBin >= 1) {	    	    
	    outer.SetPoint(nPoints,x,y);	  
	    outer.SetPointError(nPoints,exLo,exHi,eyOutLo,eyOutHi);	  
	  }
	}
	// Compute sum of weights
	double dx = exLo + exHi;
	sumWeights += y * dx;
	// Add content of overflow bin to sum of weights
	if (iBin == nBin) sumWeights += hisPtrSav[0]->GetBinContent(iBin+1) * dx;
	// Don't include underflow bin in graph	
	if (iBin == 0) continue;
	SetPoint(nPoints,x,y);
	SetPointError(nPoints,exLo,exHi,ey,ey);
      }

      // Check if normalizing to something
      if (normalizeIn && sumWeights != 0.0 && normFacIn != 0.0) {
	for (int iPoint = 0; iPoint <= nPoints; ++iPoint) {
	  double yRaw, eyRawHi, eyRawLo, x;
	  GetPoint(iPoint,x,yRaw);
	  eyRawHi = GetErrorYhigh(iPoint);
	  eyRawLo = GetErrorYlow(iPoint);
	  SetPoint(iPoint,x,normFacIn*yRaw/sumWeights);
	  SetPointEYhigh(iPoint,normFacIn*eyRawHi/sumWeights);
	  SetPointEYlow(iPoint,normFacIn*eyRawLo/sumWeights);
	  outer.GetPoint(iPoint,x,yRaw);
	  eyRawHi = outer.GetErrorYhigh(iPoint);
	  eyRawLo = outer.GetErrorYlow(iPoint);
	  outer.SetPoint(iPoint,x,normFacIn*yRaw/sumWeights);
	  outer.SetPointEYhigh(iPoint,normFacIn*eyRawHi/sumWeights);
	  outer.SetPointEYlow(iPoint,normFacIn*eyRawLo/sumWeights);
	}	
      }
      return true;
    }      
    
    // Copy color scheme from another VRgraph
    virtual void setColors(VRgraph* graph) {      
      SetMarkerSize(graph->GetMarkerSize());
      SetMarkerStyle(graph->GetMarkerStyle());
      SetMarkerColor(graph->GetMarkerColor());
      SetLineStyle(graph->GetLineStyle());
      SetLineColor(graph->GetLineColor());
      SetLineWidth(graph->GetLineWidth());
      SetFillColor(graph->GetFillColor());
      SetFillStyle(graph->GetFillStyle());
      outer.SetMarkerSize(graph->getOuter()->GetMarkerSize());
      outer.SetMarkerStyle(graph->getOuter()->GetMarkerStyle());
      outer.SetMarkerColor(graph->getOuter()->GetMarkerColor());
      outer.SetLineStyle(graph->getOuter()->GetLineStyle());
      outer.SetLineColor(graph->getOuter()->GetLineColor());
      outer.SetLineWidth(graph->getOuter()->GetLineWidth());
      outer.SetFillColor(graph->getOuter()->GetFillColor());
      outer.SetFillStyle(graph->getOuter()->GetFillStyle());
    }

    // Set colors
    virtual void setColors(int i=0) {
      // Define color schemes 
      Int_t lineCol, statCol, totCol, lineStyle, markCol, markStyle;
      double markSize(0.4), lineWidth(1.);
      // Data
      if (i == 0) {
	markCol   =  1;
	markStyle = 21;
	lineStyle =  1;
	lineCol   =  1;
	statCol    = 90;
	totCol    = 91;
	markSize  = 0.6;
	outer.SetLineColor(lineCol);
	outer.SetLineWidth(lineWidth);
	outer.SetLineStyle(1);
      } 
      else if (i == 1) {
	lineStyle =  1; 
	lineCol   = 61;
	statCol   = 64;
	totCol    = 66;
	markStyle = 20;
	markSize  = 0.4;
	markCol   = lineCol;
	lineWidth = 1.8;
      } 
      else if (i == 2) {
	lineStyle =  1; 
	lineCol   = 2;
	statCol   = 50;
	totCol    = 46;
	markCol   = lineCol;
	markStyle = 24;
	markSize  = 0.58;
	lineWidth = 1.0;
      }
      else if (i == 3) {
	lineStyle =  2; 
	lineCol   =  8;
	statCol   = 30;
	totCol    =  3;
	markCol   = lineCol;
	markStyle = 26;
	lineWidth = 1.5;
	markSize  = 0.58;
      }
      else if (i == 4) {
	lineStyle =  3;
	lineCol   = 28;
	statCol   = 26;
	totCol    = 24;
	markCol   = lineCol;
	markStyle = 23;
	lineWidth = 1.5;
	markSize  = 0.48;
      }
      else {
	lineStyle =  4;
	lineCol   = 32;
	statCol   = 30;
	totCol    = 29;
	markCol   = lineCol;
	markStyle =  27;
	lineWidth = 1.0;
	markSize  = 0.58;
      }
      SetMarkerSize(markSize);
      SetMarkerStyle(markStyle);
      SetMarkerColor(markCol);
      SetLineStyle(lineStyle);
      SetLineColor(lineCol);
      SetLineWidth(lineWidth);
      SetFillColor(statCol);
      outer.SetFillColor(totCol);
    }
    
    // Pointer to user histogram (or variation, if exists)
    TH1D* getTH1DPtr (int i=0) {
      if ( i <= int(hisPtrSav.size())-1 ) return hisPtrSav[i];
      else return NULL;
    }

  protected: 
    
    // Pointer to user histogram (vector if contains uncertainty variations)
    vector<TH1D*> hisPtrSav;
    // Container for outer uncertainty band.
    TGraphAsymmErrors outer;

    // 
    double sumWeights;
    bool hasOuterSav;
    
    // 
    bool normalizeSav;
    double normFacSav; 

    // Value determining whether bins are compatible
    // apparently, initializing this value where it is declared is not 
    // standard C++03 (but standard C++11)
    const static double TINY;

  };
  const double VRgraph::TINY = 1e-6;
  
  // ------------------------------------------------------------------------
  // class VRgraphData
  // Class for containing graphs representing experimental data points
  
  class VRgraphData : public VRgraph {

  public:

    // Constructors
    VRgraphData() { 
      hasOuterSav = true; 
    }
    VRgraphData(VRfile* dataFilePtrIn) { 
      hasOuterSav = true;
      setDataFilePtr(dataFilePtrIn);
    }
    
    // Set data file Ptr
    void setDataFilePtr(VRfile* dataFilePtrIn) {
      dataFilePtr = dataFilePtrIn;
      setColors(0);
    }

    // Update graph
    virtual bool update(bool normalizeIn=true, double normFacIn=1.0, 
                        double sigmaStat=1.0, double = 0.0) {
      if (dataFilePtr == 0) {
        cout<<"Error: cannot update dataGraph without dataFilePtr"<<endl;
        return false;
      }
      sumWeights   = dataFilePtr->getSumWeights();
      if (sumWeights == 0.0) sumWeights = 1.0;
      int nBin    = dataFilePtr->size();
      for (int iBin = 0; iBin < nBin; ++iBin) {
        double x    = dataFilePtr->getXC(iBin);
        double exLo = x - dataFilePtr->getXL(iBin);
        double exHi = dataFilePtr->getXR(iBin) - x;
        double y    = dataFilePtr->getY(iBin);
        double eyTotHi = dataFilePtr->getUtotHi(iBin);
        double eyTotLo = dataFilePtr->getUtotLo(iBin);
        double eyStatHi = dataFilePtr->getUstatHi(iBin);
        double eyStatLo = dataFilePtr->getUstatLo(iBin);	
        double eySysHi  = sqrt(max(0.0,pow(eyTotHi,2)-pow(eyStatHi,2)));
        double eySysLo  = sqrt(max(0.0,pow(eyTotLo,2)-pow(eyStatLo,2)));
        // Scale stat error by sigmaStat (to show nSigma contours)
        eyStatHi *= sigmaStat;
        eyStatLo *= sigmaStat;
        eyTotHi = sqrt(pow(eyStatHi,2) + pow(eySysHi,2));
        eyTotLo = sqrt(pow(eyStatLo,2) + pow(eySysLo,2));
        SetPoint(iBin,x,y);
        SetPointError(iBin,exLo,exHi,eyStatLo,eyStatHi);
        outer.SetPoint(iBin,x,y);	  
        outer.SetPointError(iBin,exLo,exHi,eyTotLo,eyTotHi);	  
      }
      
      // Check if normalizing to something
      if (normalizeIn && sumWeights != 0.0 && normFacIn != 0.0) {
        for (int iBin = 0; iBin < nBin; ++iBin) {
          double yRaw, eyRawHi, eyRawLo, x;
          GetPoint(iBin,x,yRaw);
          eyRawHi = GetErrorYhigh(iBin);
          eyRawLo = GetErrorYlow(iBin);
          SetPoint(iBin,x,normFacIn*yRaw/sumWeights);
          SetPointEYhigh(iBin,normFacIn*eyRawHi/sumWeights);
          SetPointEYlow(iBin,normFacIn*eyRawLo/sumWeights);
          eyRawHi = outer.GetErrorYhigh(iBin);
          eyRawLo = outer.GetErrorYlow(iBin);
          outer.SetPoint(iBin,x,normFacIn*yRaw/sumWeights);
          outer.SetPointEYhigh(iBin,normFacIn*eyRawHi/sumWeights);
          outer.SetPointEYlow(iBin,normFacIn*eyRawLo/sumWeights);
        }
      }
      return true;
    }

  protected: 
    
    // Pointer to user histogram (vector if contains uncertainty variations)
    VRfile* dataFilePtr;

    // Format specifier
    vector<int> format;

  };

  // ------------------------------------------------------------------------
  // class VRgraphNorm
  // Class for ratios of histograms (e.g., normalized to data)

  class VRgraphNorm : public VRgraph {

  public:
    
    // Constructors
    VRgraphNorm() { 
      hasOuterSav = false; 
    }
    VRgraphNorm(VRgraph* numeratorIn, VRgraph* denominatorIn) {
      hasOuterSav = false;
      setNumerator(numeratorIn);
      setDenominator(denominatorIn);
    }

    // Functions to set
    void setNumerator(VRgraph* graphIn) {
      numeratorPtr = graphIn;
      if (numeratorPtr) {
	hasNumerator = true;
	hasOuterSav  = numeratorPtr->hasOuter();
      }
    }
    void setDenominator(VRgraph* graphIn) {
      denominatorPtr = graphIn;
      if (denominatorPtr) hasDenominator = true;
    }
    
    // Update graph
    virtual bool update(double sigmaIn=1.0) {

      // Sanity checks
      if (!hasNumerator || !hasDenominator) return false;
      if (numeratorPtr->GetN() <= 0) return false; 
      if (denominatorPtr->GetN() <= 0) return false; 
      
      // 
      hasOuterSav = numeratorPtr->hasOuter();
      
      // Compute normalized histogram (using binning from denominator)
      int nBin    = denominatorPtr->GetN();
      int nPoints = -1;
      int iFirst  = 0;
      int iLast   = nBin -1;
      for (int iBin = iFirst; iBin <= iLast; ++iBin) {
        
        // Extract bin to use for normalization
        double xC, yDenom, exLo, exHi, xL, xR;
        denominatorPtr->GetPoint(iBin,xC,yDenom);	
        // Sanity check
        if (yDenom == 0.0) continue;
        exLo = denominatorPtr->GetErrorXlow(iBin);
        exHi = denominatorPtr->GetErrorXhigh(iBin);
        xL = xC - exLo;
        xR = xC + exHi;
        
        // Loop over numerator bins and rebin like denominator
        double sumw     = 0.0;
        double sumy     = 0.0;
        // Statistical MC generation errors
        double sumErr2  = 0.0;
        // Note: up/down variation curves must be added linearly!
        double sumUpper = 0.0;
        double sumLower = 0.0;
        VRgraph* h=numeratorPtr;
        TGraphAsymmErrors* hU=numeratorPtr->getOuter();
        int nh = h->GetN();
        for (int hBin=0; hBin<nh; hBin++){
          // Get central values for this numerator bin
          double xNum, yNum, xNumU, yNumU;
          h->GetPoint(hBin,xNum,yNum);
          hU->GetPoint(hBin,xNumU,yNumU);	  
          double xLnum = xNum - h->GetErrorXlow(hBin);
          double xRnum = xNum + h->GetErrorXhigh(hBin);
          double bw = xRnum-xLnum;
          if (xRnum <= xL) continue;
          if (xLnum >= xR) continue;
          if (bw <= 0.0) continue;
          double w = 1.0;
          w -= max(0.,xL-xLnum)/bw;
          w -= max(0.,xRnum-xR)/bw;
          w = max(w,0.0);
          sumw += w;
          double uStat   = 0.5*(h->GetErrorYlow(hBin)+h->GetErrorYhigh(hBin));
          double ySysHi  = yNumU + sqrt(max(0.0,pow(hU->GetErrorYhigh(hBin),2) 
                                            - pow(uStat,2)));
          double ySysLo  = yNumU - sqrt(max(0.0,pow(hU->GetErrorYlow(hBin),2) 
                                            - pow(uStat,2)));
          sumy     += w*yNum;	    
          sumErr2  += w*pow(uStat,2);
          sumUpper += w*ySysHi;
          sumLower += w*ySysLo;
        }	  	  
        // Sanity check
        if (sumw <= 0.) continue;
        
        // Compute average values and total stat error
        double yTh   = sumy/sumw;	  
        double eyThStat  = sigmaIn * sqrt(sumErr2)/sumw;
        double yVarUpper = sumUpper/sumw;
        double yVarLower = sumLower/sumw;
        double yVarTh    = 0.5*(yVarUpper+yVarLower);
        double eyVarTh   = yVarUpper-yVarTh;
        
        // Normalize by data
        Double_t yThN     = yTh/yDenom;
        Double_t eyThN    = eyThStat/yDenom;	  
        Double_t yVarThN  = yVarTh/yDenom;
        // Add stat error in quadrature for normalized total uncertainty
        Double_t eyVarThN = sqrt(pow(eyVarTh,2) + pow(eyThStat,2))/yDenom;	
        
        // Limit to max +/- 50% (to clip plot
        double eyThNhi = (yThN + eyThN > 1.5) ? max(0.0,1.5-yThN) : eyThN;
        double eyThNlo = (yThN - eyThN < 0.5) ? max(0.0,yThN-0.5) : eyThN;
        double eyVarThNhi = (yVarThN + eyVarThN > 1.5) ? max(0.0,1.5-yVarThN)
        : eyVarThN;
        double eyVarThNlo = (yVarThN - eyVarThN < 0.5) ? max(0.0,yVarThN-0.5)
        : eyVarThN;
        
        // Set point value and errors;
        ++nPoints;
        SetPoint(nPoints,xC,yThN);
        SetPointError(nPoints,exLo,exHi,eyThNlo,eyThNhi);
        outer.SetPoint(nPoints,xC,yVarThN);	  
        outer.SetPointError(nPoints,exLo,exHi,eyVarThNlo,eyVarThNhi);	  
      }
      return true;
    }
    
  protected:

    VRgraph* numeratorPtr;
    VRgraph* denominatorPtr;
    bool hasNumerator;
    bool hasDenominator;

  };


  // ------------------------------------------------------------------------
  // class VRgraphVar
  // Class for variations of histograms (e.g., relative uncertainties)

  class VRgraphVar : public VRgraph {

  public:
    
    // Constructors
    VRgraphVar() { 
      hasOuterSav = false; 
    }
    VRgraphVar(VRgraph* referenceIn,int iVarIn =1) {
      hasOuterSav = referenceIn->hasOuter();
      iVarSav     = min(referenceIn->nOuter()-1,max(0,iVarIn));      
      setReference(referenceIn);
    }

    // Functions to set
    void setReference(VRgraph* graphIn) {
      referencePtr = graphIn;
      if (referencePtr) {
	hasReference = true;
	if (referencePtr->hasOuter()) hasOuterSav = true;
      }
    }
    
    // Update graph
    virtual bool update(bool = true, double = 0.0, double = 0.0, double = 0.0) {

      // Sanity checks
      if (referencePtr->GetN() <= 0) return false; 

      // Compute difference histogram (using binning from reference)
      // and normalize by reference
      TH1D* hRef  = referencePtr->getTH1DPtr(0);
      TH1D* hVar  = referencePtr->getTH1DPtr(iVarSav);
      int nBin    = hRef->GetNbinsX();
      int nPoints = -1;
      int iFirst  = 1;
      int iLast   = nBin;
      // Loop over histogram bins
      for (int iBin = iFirst; iBin <= iLast; ++iBin) {
	double xRef = hRef->GetBinCenter(iBin);
	double yRef = hRef->GetBinContent(iBin);
	//double eRef = hRef->GetBinError(iBin);
	double yVar = hVar->GetBinContent(iBin);
	//double eVar = hVar->GetBinError(iBin);
	double yMin = min(yRef,yVar);
	double yMax = max(yRef,yVar);
	for (int iVar=1; iVar<int(referencePtr->nOuter()); ++iVar) {
	  if (iVar == iVarSav) continue;
	  double yTmp = referencePtr->getTH1DPtr(iVar)->GetBinContent(iBin);
	  if (yTmp > yMax) yMax = yTmp;
	  else if (yTmp < yMin) yMin = yTmp;
	}
	// Sanity check (require non-zero total uncertainty)
	if (yMax - yMin == 0.0) continue;

	// Is this an up or down variation?
	double yFrac = 0.;
	double eyFracHi = 0.;
	double eyFracLo = 0.;
	if (yVar >= yRef) {
	  // Sanity check (require non-zero total up variation)
	  if (yMax - yRef <= 0.0) continue;
	  yFrac = (yVar - yRef)/(yMax - yRef);	  
	  //eyFracHi = sqrt(pow(eVar,2)+pow(eRef,2))/(yMax - yRef);
	  //if (yFrac + eyFracHi > 1.0) eyFracHi = 1.0 - yFrac;
	  eyFracLo = abs(yFrac);
	} else {
	  if (yRef - yMin <= 0.0) continue;
	  yFrac = (yVar - yRef)/(yRef - yMin);
	  //eyFracLo = sqrt(pow(eVar,2)+pow(eRef,2))/(yRef - yMin);
	  //if (yFrac - eyFracLo < -1.0) eyFracLo = -1.0 - yFrac;
	  eyFracHi = abs(yFrac);
	}

	// Set point value and errors;
	++nPoints;
	SetPoint(nPoints,xRef,yFrac);
	SetPointError(nPoints,0.,0.,eyFracLo,eyFracHi);
      }
      return true;
    }
    
    void setStyle(int iVar=1) {
      if (iVar == 1) {
	SetLineStyle(1);
	SetLineWidth(1.0);	
	SetFillColor(kBlack);
      }
      else if (iVar == 2) {
	SetLineStyle(1);
	SetLineWidth(1.0);	
	SetLineColor(kBlue-10);
	SetFillColor(kBlue-10);
      }
      else if (iVar == 3) {
	SetLineStyle(1);
	SetLineWidth(1.0);	
	SetLineColor(kRed-10);
	SetFillColor(kRed-10);
      }
      else if (iVar == 4) {
	SetLineStyle(1);
	SetLineWidth(2.0);	
	SetLineColor(kBlue-9);
	SetFillColor(kBlue-9);
	SetFillStyle(3305);
      }
      else if (iVar == 5) {
	SetLineStyle(1);
	SetLineWidth(2.0);	
	SetLineColor(kRed-8);
	SetFillColor(kRed-8);
	SetFillStyle(3305);
      }
      else if (iVar == 6) {
	SetLineStyle(1);
	SetLineWidth(2.0);	
	SetLineColor(kBlue-7);
	SetFillColor(kBlue-7);
	SetFillStyle(0);
      }
      else if (iVar == 7) {
	SetLineStyle(1);
	SetLineWidth(2.0);	
	SetLineColor(kRed-6);
	SetFillColor(kRed-6);
	SetFillStyle(0);
      }
      else if (iVar == 8) {
	SetLineStyle(1);
	SetLineWidth(1.0);	
	SetLineColor(kGray+2);
	SetFillStyle(0);
      }
      else if (iVar == 9) {
	SetLineStyle(1);
	SetLineWidth(1.0);	
	SetLineColor(kGray+3);
	SetFillStyle(0);
      }
      else if (iVar == 10) {
	SetLineStyle(3);
	SetLineWidth(1.0);	
	SetLineColor(kBlue+2);
	SetFillColor(kBlue+2);
	SetFillStyle(0);
      }
      else if (iVar == 11) {
	SetLineStyle(3);
	SetLineWidth(2.0);	
	SetLineColor(kRed+2);
	SetFillColor(kRed+2);
	SetFillStyle(0);
      }
    }


  protected:

    VRgraph* referencePtr;
    bool hasReference;

  private:
    int iVarSav;

  };


  // ------------------------------------------------------------------------
  // class VRdisplay
  // Contains the base class for a single 1D plot

  class VRdisplay {
    
  public:

    // Constructor
    VRdisplay(string displayNameIn, int xSizIn = 300, int ySizIn = 400,
	      int xOffsetIn=-1, int yOffsetIn=-1) 
      : nPad(1), isInitPad(false), hasAxesSav(false), hasDataSav(false), 
      hasTitle(false), hasLegend(false), hasMCnames(false), hasDataRef(false),
      hasSignature(false) {
      displayNameSav   = displayNameIn;
      displayCanvasPtr = new TCanvas(displayNameIn.c_str(),
				     displayNameIn.c_str(), 
				     xOffsetIn, yOffsetIn,xSizIn,ySizIn);      
    }
  
    // Init pointers and set settings
    void init(VRsettings* settingsPtrIn) {
      settingsPtr   = settingsPtrIn;
      logY          = settingsPtr->logY;
      normalize     = settingsPtr->normalize;
      normFac       = settingsPtr->normFac;
      optStat       = settingsPtr->optStat; 
      verbose       = settingsPtr->verbose;
      sigmaStat     = settingsPtr->sigmaStat;
      nPad          = 1;      
      setShowRatio(settingsPtr->showRatioPad);
      setShowRelUnc(settingsPtr->showRelUncPads);
      setMCnames(settingsPtr->MCnames);
      setLegend(settingsPtr->legend);
      setSignature(settingsPtr->signature);
      setLeftLegend(false);
    }

    // Functions to Get    
    bool getLogY() { return logY; }
    bool getLinY() { return !logY; }
    bool getNormalize() { return normalize; }
    int  getOptStat() { return optStat; }
    double getSigmaMC() { return sigmaStat; }    
    double getNormFac() { return normFac; }
    string getName() { return displayNameSav; }

    // Functions to check whether axes and data have been defined
    bool hasAxes() {return hasAxesSav;}
    bool hasData() {return hasDataSav;}    

    // Add pad (for ratio to data or decomposition of vector of histos)
    void addPad() {
      nPad++;
      // When adding many pads, increase label font sizes
      for (int i=0; i<nPad-1; ++i) {
	double labSizeX = 1.07 * axesPtr->GetXaxis()->GetLabelSize();
	double labSizeY = 1.07 * axesPtr->GetYaxis()->GetLabelSize();
	axesPtr->GetXaxis()->SetLabelSize(labSizeX);
	axesPtr->GetYaxis()->SetLabelSize(labSizeY);      
	double titSizeX = 1.09 * axesPtr->GetXaxis()->GetTitleSize();
	double titSizeY = 1.09 * axesPtr->GetYaxis()->GetTitleSize();
	axesPtr->GetXaxis()->SetTitleSize(titSizeX);
	axesPtr->GetYaxis()->SetTitleSize(titSizeY);      
	double titOffsetY = 0.94*axesPtr->GetYaxis()->GetTitleOffset();
	axesPtr->GetYaxis()->SetTitleOffset(titOffsetY);      
	// Ratio pad
	double labNormSizeX = 1.05 * axesNormPtr->GetXaxis()->GetLabelSize();
	double labNormSizeY = 1.05 * axesNormPtr->GetYaxis()->GetLabelSize();
	axesNormPtr->GetXaxis()->SetLabelSize(labNormSizeX);
	axesNormPtr->GetYaxis()->SetLabelSize(labNormSizeY);      
	double titNormSizeX = 1.18*axesNormPtr->GetXaxis()->GetTitleSize();
	double titNormSizeY = 1.10*axesNormPtr->GetYaxis()->GetTitleSize();
	axesNormPtr->GetXaxis()->SetTitleSize(titNormSizeX);
	axesNormPtr->GetYaxis()->SetTitleSize(titNormSizeY);      
	double titNormOffsetY = 0.9*axesNormPtr->GetYaxis()->GetTitleOffset();
	axesNormPtr->GetYaxis()->SetTitleOffset(titNormOffsetY);      
      }
      // RelUnc pad(s)
      if (nPad >= 3) {
	double labVarSizeX = 1.75 * axesVarPtr->GetXaxis()->GetLabelSize();
	double labVarSizeY = 1.75 * axesVarPtr->GetYaxis()->GetLabelSize();
	axesVarPtr->GetXaxis()->SetLabelSize(labVarSizeX);
	axesVarPtr->GetYaxis()->SetLabelSize(labVarSizeY);      
	double titVarSizeX = max(1.75 * axesVarPtr->GetXaxis()->GetTitleSize(),
      			 double(axesVarPtr->GetXaxis()->GetLabelSize()));
	double titVarSizeY = 1.85 * axesVarPtr->GetYaxis()->GetTitleSize();
	double titVarOffsetY = 0.6 * axesVarPtr->GetYaxis()->GetTitleOffset();
	axesVarPtr->GetXaxis()->SetTitleSize(titVarSizeX);
	axesVarPtr->GetYaxis()->SetTitleSize(titVarSizeY);      
	axesVarPtr->GetYaxis()->SetTitleOffset(titVarOffsetY);
	isInitPad = false;
      }
    }

    // Goto pad
    void gotoPad(int iPadIn = 1) {      
      displayCanvasPtr->cd();      

      // If pads already defined, just cd to the desired one
      if (isInitPad) {
	gPad->cd(min(iPadIn,nPad));
	return;
      }

      // If pads not yet defined, create them 
      gPad->Divide(1,nPad);
      double nDiv = 4. + nPad;
      double topMargin    = 0.5*gPad->GetTopMargin();
      double bottomMargin = gPad->GetBottomMargin();
      if (nPad >= 2) bottomMargin *= 1.75 + 0.5*nPad;
      double leftMargin   = (1.05+nPad/10.)*gPad->GetLeftMargin();
      double rightMargin  = 0.5*gPad->GetRightMargin();      
      for (int iPad=1; iPad<=nPad; ++iPad) {
	displayCanvasPtr->cd();	
	gPad->cd(iPad);
	double yHi = 1.0;
	if (iPad >= 2) yHi = 1.0 - double(2+iPad)/nDiv;
	double yLo = 1.0 - double( (iPad+3)/nDiv );
	if (iPad == nPad) yLo = 0.0;
	// Set pad coordinates
	gPad->SetPad(0.0,yLo,1.0,yHi);
	// Set inner margins
	if (iPad == 1) gPad->SetTopMargin(topMargin);
	else gPad->SetTopMargin(0.0);
	if (iPad == nPad) gPad->SetBottomMargin(bottomMargin);
	else gPad->SetBottomMargin(0.0);
	gPad->SetLeftMargin(leftMargin);
	gPad->SetRightMargin(rightMargin);
	if (!isInitPad) {
	  gPad->Draw();
	} else {
	  gPad->Modified();
	}
      }
      isInitPad = true;
      gPad->cd(min(iPadIn,nPad));
    }

    // Functions to Set
    void setVerbose(int verboseIn) {verbose = verboseIn;};
    void setOptStat(int optStatIn) {optStat = optStatIn;}
    void setSigmaMC(double sigmaStatIn) {sigmaStat = sigmaStatIn;};
    void setNormalize(bool normalizeIn = true, double normFacIn = 1.0) {
      normalize  = normalizeIn;
      normFac    = normFacIn;
    }
    void setLogY(bool logYin = true) { 
      if (logY != logYin) logY = logYin;
      displayCanvasPtr->cd(1);	
      if (logYin) gPad->SetLogy(); 
      else gPad->SetLogy(0);
    }
    void setLinY(bool linYin = true) { setLogY(!linYin); }
    void setSignature(string signatureIn="V I N C I A R O O T") {
      signature = signatureIn;
    }
    void setLeftLegend(bool leftLegendIn=true) { leftLegend = leftLegendIn; } 
    void setShowRatio(bool showRatioIn=true) { 
      showRatioPad = showRatioIn; 
    }
    void setShowRelUnc(bool showRelUncIn=true) {
      showRelUncPads = showRelUncIn;
    }

    // Check if axes and/or data has been defined
    void setAxes(TH1D* initHis) {
      if (hasAxes()) {
	delete axesPtr; 
	delete axesNormPtr;
	delete axesVarPtr;
      }
      axesPtr = new TH1D(*initHis);
      axesPtr->Reset("ICES");
      hasAxesSav = true;
      // Axis histogram is pure dummy, has no stats, so switch off stats display
      axesPtr->SetStats(false);
      // Set fonts and axis tick divisions 
      axesPtr->GetXaxis()->SetLabelFont(42);
      axesPtr->GetYaxis()->SetLabelFont(42);
      axesPtr->GetXaxis()->SetNdivisions(505);
      axesPtr->GetYaxis()->SetNdivisions(606);
      axesPtr->GetXaxis()->SetTitleFont(42);
      double labSizeX = 1.2 * axesPtr->GetXaxis()->GetLabelSize();
      double labSizeY = 1.2 * axesPtr->GetYaxis()->GetLabelSize();
      axesPtr->GetXaxis()->SetLabelSize(labSizeX);
      axesPtr->GetYaxis()->SetLabelSize(labSizeY);
      axesPtr->GetYaxis()->SetTitleSize(1.1*axesPtr->GetYaxis()->GetTitleSize());
      axesPtr->GetYaxis()->SetTitleOffset(1.18*axesPtr->GetYaxis()->GetTitleOffset());
      axesPtr->GetYaxis()->SetTitleFont(42);
      // Switch off lame ROOT title
      axesPtr->SetTitle("");

      // Set axis histo for ratio pad
      axesNormPtr = new TH1D(*axesPtr);
      axesNormPtr->SetMinimum(0.5);
      axesNormPtr->SetMaximum(1.5);
      // X axis
      axesNormPtr->GetXaxis()->SetTickLength(0.1);
      axesNormPtr->GetXaxis()->SetLabelSize(1.9*axesPtr->GetXaxis()->GetLabelSize());
      axesNormPtr->GetXaxis()->SetTitleFont(42);
      axesNormPtr->GetXaxis()->SetTitleSize(1.85*axesPtr->GetXaxis()->GetTitleSize());
      axesNormPtr->GetXaxis()->SetTitleOffset(
        1.*axesPtr->GetXaxis()->GetTitleOffset());
      // Y axis
      axesNormPtr->GetYaxis()->SetLabelFont(42);
      axesNormPtr->GetYaxis()->SetLabelSize(1.8*axesPtr->GetYaxis()->GetLabelSize());
      axesNormPtr->GetYaxis()->SetLabelOffset(0.01);
      axesNormPtr->GetYaxis()->SetTitleFont(42);
      axesNormPtr->GetYaxis()->SetTitleSize(1.8*axesPtr->GetYaxis()->GetTitleSize());
      axesNormPtr->GetYaxis()->SetTitle("Theory/Data");	  
      axesNormPtr->GetYaxis()->SetTitleOffset(0.63*axesPtr->GetYaxis()->GetTitleOffset());
      axesNormPtr->GetYaxis()->CenterTitle();
      axesNormPtr->GetYaxis()->SetTickLength(0.025);

      // Set axis histo for relUnc pads (uncertainty variations)
      axesVarPtr = new TH1D(*axesPtr);
      axesVarPtr->SetMinimum(-1.25);
      axesVarPtr->SetMaximum(1.25);
      // X axis
      axesVarPtr->GetXaxis()->SetTickLength(2.*axesPtr->GetXaxis()->GetTickLength());
      axesVarPtr->GetXaxis()->SetLabelOffset(2.*axesPtr->GetXaxis()->GetLabelOffset());
      axesVarPtr->GetXaxis()->SetLabelSize(2.*axesPtr->GetXaxis()->GetLabelSize());
      axesVarPtr->GetXaxis()->SetTitleFont(42);
      axesVarPtr->GetXaxis()->SetTitleSize(2.5*axesPtr->GetXaxis()->GetTitleSize());
      axesVarPtr->GetXaxis()->SetTitleOffset(
        1.*axesPtr->GetXaxis()->GetTitleOffset());
      // Y axis
      axesVarPtr->GetYaxis()->SetNdivisions(203);
      axesVarPtr->GetYaxis()->SetTickLength(axesPtr->GetYaxis()->GetTickLength());
      axesVarPtr->GetYaxis()->SetLabelFont(42);
      axesVarPtr->GetYaxis()->SetLabelSize(1.8*axesPtr->GetYaxis()->GetLabelSize());
      axesVarPtr->GetYaxis()->SetLabelOffset(1.8*axesPtr->GetYaxis()->GetLabelOffset());
      axesVarPtr->GetYaxis()->SetTitleFont(42);
      axesVarPtr->GetYaxis()->SetTitle("Rel. Unc.");	  
      axesVarPtr->GetYaxis()->SetTitleSize(2*axesPtr->GetYaxis()->GetTitleSize());
      axesVarPtr->GetYaxis()->SetTitleOffset(0.5*axesPtr->GetYaxis()->GetTitleOffset());
      axesVarPtr->GetYaxis()->CenterTitle();
    }

    //------------------------------------------------------------------
    // VRdisplay: TH1D I/O

    // Add a histogram, not necessarily the first
    void addTH1DPtr(TH1D* hisPtrIn) {
      if (!hasAxes()) {
	setAxes(hisPtrIn);
	graphs.resize(0);
      }
      graphs.push_back(VRgraph(hisPtrIn));      
      // Default: assign color set n to graph number n
      graphs[graphs.size()-1].setColors(graphs.size());
    }

    // Add a histogram (with vector uncertainties), not necessarily the first
    void addTH1DPtr(vector<TH1D*> hisPtrIn) {
      if (!hasAxes()) {
        setAxes(hisPtrIn[0]);
        graphs.resize(0);
      }
      graphs.push_back(VRgraph(hisPtrIn));
      graphs[graphs.size()-1].setColors(graphs.size());
      
      // Add pad for decomposition of vector band
      if (hisPtrIn.size() >= 2 && showRelUncPads) addPad();
    }
    
    // Get histogram pointer corresponding to graph i, variation j
    TH1D* getTH1DPtr(int i=0, int j=0) {
      if ( i <= int(graphs.size())-1 ) return graphs[i].getTH1DPtr(j);
      else return NULL;
    }

    //========================================================================
    // VRdisplay: FILE I/O

    //------------------------------------------------------------------------
    // getFormat(): 
    // Function to convert a list of integers to vector<int> 
    void getFormat(vector<int>& format, int i1, int i2, int i3=-1, int i4=-1,
		   int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, 
		   int i10=-1, int i11=-1, int i12=-1) {
      // Store file format 
      format.push_back(i1);
      format.push_back(i2); 
      if (i3 >= 0) format.push_back(i3); else return;
      if (i4 >= 0) format.push_back(i4); else return;
      if (i5 >= 0) format.push_back(i5); else return;
      if (i6 >= 0) format.push_back(i6); else return;
      if (i7 >= 0) format.push_back(i7); else return;
      if (i8 >= 0) format.push_back(i8); else return;
      if (i9 >= 0) format.push_back(i9); else return;
      if (i10 >= 0) format.push_back(i10); else return;
      if (i11 >= 0) format.push_back(i11); else return;
      if (i12 >= 0) format.push_back(i12); else return;
    }

    //------------------------------------------------------------------------
    // setDataFile():
    // Set file to be treated as experimental data 
    bool setDataFile(string dataFileName, bool hasSysErrors = true) {
      if (hasSysErrors) return setDataFile(dataFileName,10,20,30,40,51,52,61,62);
      else return setDataFile(dataFileName,10,20,30,40,51,52);
    }
    bool setDataFile(string dataFileName, int i1, int i2, int i3=-1, int i4=-1, 
		     int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, 
		     int i10=-1, int i11=-1, int i12=-1) {     

      // Verbose output
      if (verbose >= 3) cout<<"(VRdisplay): adding dataFile "<<dataFileName
			    <<" to display "<<displayNameSav<<endl;

      // Create new dataFilePtr with specified format
      vector<int> format;
      getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12);
      dataFilePtr = new VRfile(format);

      // Read file 
      bool ok = dataFilePtr->readFile(dataFileName, verbose);
      if (ok) {
	
	// Create new VRgraph to contain data points
	dataPtr = new VRgraphData(dataFilePtr);
	dataNormPtr = new VRgraphNorm(dataPtr,dataPtr);
	dataNormPtr->setColors(0);

	// This display has data
	hasDataSav = true;
	if (showRatioPad) addPad();

      } else {
	cout<<"VRDisplay::setDataFile() error reading "<<dataFileName<<endl;
      }

      // Return
      return ok;
    }

    //------------------------------------------------------------------------
    // addHisFile() : 
    // Add file with saved histogram
    bool addHisFile(string hisFileName, bool hasSysErrors = true) {
      if (hasSysErrors) return addHisFile(hisFileName,10,20,30,40,51,52,61,62);
      else return addHisFile(hisFileName,10,20,30,40,51,52);
    }
    bool addHisFile(string hisFileName, int i1, int i2, int i3=-1, int i4=-1, 
		    int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, 
		    int i10=-1, int i11=-1, int i12=-1) {     

      // Verbose output
      if (verbose >= 3) cout<<"(VRdisplay): adding histogram file "<<hisFileName
			    <<" to display "<<displayNameSav<<endl;

      // Add to list of hisFilePtrs.
      int nHisFiles = hisFilePtrs.size();
      hisFilePtrs.resize(nHisFiles+1);

      // Create new dataFilePtr with specified format
      vector<int> format;
      getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12);
      hisFilePtrs[nHisFiles] = new VRfile(format);

      // Read file and return success/failure
      bool ok = hisFilePtrs[nHisFiles]->readFile(hisFileName, verbose);
      return ok;
    }

    //------------------------------------------------------------------------
    // update(): 
    // Update Display 
    void update() {
      
      // CD to this display
      displayCanvasPtr->cd();	
      
      // If several pads, first do main plot
      gotoPad(1);
      
      // Update and (re)draw axis
      if (!hasAxes()) {
        if (graphs.size() >= 1) setAxes(graphs[0].getTH1DPtr());
        else return;
      }
      
      // Update max/min if possible.
      if (graphs.size() >= 1) {
        TH1D* hisPtr = graphs[0].getTH1DPtr();
        if (hisPtr != NULL) {
          // Check if min/max changed by user
          double yMinSav = axesPtr->GetMinimum();
          double yMaxSav = axesPtr->GetMaximum();
          double yMinNow = hisPtr->GetMinimum();
          double yMaxNow = hisPtr->GetMaximum();
          if (yMinSav != yMinNow || yMaxSav != yMaxNow) {
            axesPtr->SetMinimum(yMinNow);
            axesPtr->SetMaximum(yMaxNow);
          }
          
          // Also update log Y axis properties (force linear if yMin <= 0.0)
          if ( yMinNow <= 0.0 && gPad->GetLogy() ) setLogY(false); 
          else if ( yMinNow > 0.0 && logY && !gPad->GetLogy() ) setLogY(true);
          
          // Update title 
          setTitle();
        }
      }
      
      // Draw axes
      if (nPad >= 2) {
        // Switch off X axis title for main histo
        axesPtr->GetXaxis()->SetTitle("");
        // Draw relUnc pads
        if (showRelUncPads) {
          int iPad = 1;
          for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
            if (graphs[iGraph].hasOuter()) {
              ++iPad;
              displayCanvasPtr->cd(iPad);
              axesVarPtr->Draw();
            }
          }
        }
        // Draw ratio pad
        if (showRatioPad && hasDataSav && dataPtr) {
          displayCanvasPtr->cd(nPad);
          axesNormPtr->Draw();	
        }
      }
      // Draw main display
      displayCanvasPtr->cd(1);
      axesPtr->Draw();      
      
      // Update data
      if (hasDataSav && dataPtr) {
        displayCanvasPtr->cd(1);
        dataPtr->update(normalize,normFac,sigmaStat);
        dataPtr->DrawOuter("E0Z");
        dataPtr->Draw("PE1");
        if (dataNormPtr && nPad >= 2 && showRatioPad) {
          displayCanvasPtr->cd(nPad);
          dataNormPtr->update();
          dataNormPtr->DrawOuter("e2");
          dataNormPtr->Draw("e2");
          axesNormPtr->Draw("sameaxis");
          //gPad->RedrawAxis();
        }
      }
      
      // Update graphs
      displayCanvasPtr->cd(1);
      for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
        
        // Set X scatter
        double scatterX = 0.1*pow(-1.0,iGraph*1.0) * iGraph/graphs.size();
        
        // Skip if update fails
        if (!graphs[iGraph].update(normalize,normFac,sigmaStat,scatterX)) continue; 
        
        // Outer band
        if (graphs[iGraph].hasOuter()) {
          int fillStyle = 3354;
          if (iGraph == 1) fillStyle = 3345;	    
          else if (iGraph >= 2) fillStyle = 3307;
          graphs[iGraph].getOuter()->SetFillStyle(fillStyle);	    	
        }
        
        // Draw graph
        graphs[iGraph].DrawOuter("e2");    
        graphs[iGraph].Draw("line P e Z");    
      }
      
      // Update normalized graphs
      bool hasRatioPad = false;
      if (dataNormPtr && nPad >=2 && hasDataSav && showRatioPad) { 	  
        hasRatioPad = true;
        displayCanvasPtr->cd(nPad);
        if (graphsNorm.size() < graphs.size()) {
          graphsNorm.resize(0);
          for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
            graphsNorm.push_back(VRgraphNorm(&graphs[iGraph],dataPtr));	    
            graphsNorm[iGraph].setColors(&graphs[iGraph]);
            if (graphsNorm[iGraph].hasOuter()) {
              int fillStyle = 3354;
              if (iGraph == 1) fillStyle = 3345;	    
              else if (iGraph >= 2) fillStyle = 3307;
              graphsNorm[iGraph].SetFillStyle(fillStyle);	    
            }
          }
        }
        for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
          // Update
          graphsNorm[iGraph].update(sigmaStat);
          // Skip if no points to be plotted
          if (graphsNorm[iGraph].GetN() <= 0) continue;
          // Draw graph
          graphsNorm[iGraph].DrawOuter("e2");
          graphsNorm[iGraph].Draw("linePeZ");
        }
      }
      
      // Update relative uncertainty variation pads
      int nUncVarPads = nPad - hasRatioPad -1;
      if (nUncVarPads >= 1) {
        int iPad = 1;
        for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
          // Only do RelUnc pads for graphs that have (>1) outer bands
          if (!graphs[iGraph].hasOuter() || graphs[iGraph].nOuter() <= 1) 
            continue;
          ++iPad;
          // Sanity check: allow for max 3 separate RelUnc pads
          if (iPad > nPad || iPad >= 3) continue;
          displayCanvasPtr->cd(iPad);
          // Create vector if not already done
          if (graphsVar.size() <= 1) {
            graphsVar.resize(0);
            for (int iVar=1; iVar<graphs[iGraph].nOuter(); ++iVar) {
              graphsVar.push_back(VRgraphVar(&graphs[iGraph],iVar));
              graphsVar[graphsVar.size()-1].setStyle(iVar);
            }
          }
          for (int iVar=0; iVar<int(graphsVar.size()); ++iVar) {
            graphsVar[iVar].update();
            // Skip if no points to be plotted
            if (graphsVar[iVar].GetN() <= 0) continue;
            // Draw graph
            graphsVar[iVar].Draw("e3line");
          }
        }
      }
      
      // Draw title
      displayCanvasPtr->cd(1);      
      if (hasTitle) titleTexPtr->Draw();
      
      // Draw MC names
      displayCanvasPtr->cd(1);
      if (!hasMCnames) setMCnames(settingsPtr->MCnames);
      if (hasMCnames) MCnamesPtr->Draw();
      
      // Draw data ref
      displayCanvasPtr->cd(1);
      if (hasDataSav) {
        if (!hasDataRef) setDataRef(dataFilePtr->getREF());
        if (hasDataRef) dataRefPtr->Draw();
      }
      
      // Draw legend (so it appears on top of curves)
      displayCanvasPtr->cd(1);
      if (!hasLegend) setLegend(settingsPtr->legend);
      if (hasLegend) legendPtr->Draw();
      
      // Draw signature
      if (!hasSignature) {
        double xR = 1.-0.9*gPad->GetRightMargin();      
        double yR = gPad->GetBottomMargin();
        if (nPad >= 2) yR = 0.;
        signaturePtr = new TLatex(xR,yR,signature.c_str());
        signaturePtr->SetNDC(true);
        signaturePtr->SetTextAngle(90);
        signaturePtr->SetTextAlign(13);
        signaturePtr->SetTextColor(12);
        signaturePtr->SetTextFont(42);
        double sigSize = 0.02 + 0.003*nPad;
        signaturePtr->SetTextSize(sigSize);
        hasSignature = true;
      }
      signaturePtr->Draw();
      
      // Then update canvas
      displayCanvasPtr->cd();
      displayCanvasPtr->Update();      
    }
    
    //------------------------------------------------------------------------
    // Save display image
    void saveDisplay(string prefix="none",  
		     string fileType="png",string option="Landscape") {
      // Default file name is the name of display
      string fileName = displayNameSav;
      // If specific prefix given, append display name to that 
      if (prefix != "none") fileName = prefix+"-"+displayNameSav;	  
      // First loop for specific file types
      if (fileType.find("eps") != string::npos) fileName+=".eps";
      else if (fileType.find("ps") != string::npos) fileName+=".ps";
      else if (fileType.find("gif") != string::npos) fileName+=".gif";
      else if (fileType.find("jpg") != string::npos) fileName+=".jpg";
      else if (fileType.find("png") != string::npos) fileName+=".png";
      else if (fileType.find("pdf") != string::npos) fileName+=".pdf";
      // Then try generic extension. Don't add dot if ext already has one.
      else if (fileType.find(".") != string::npos) fileName+=fileType;
      else fileName+="."+fileType;
      displayCanvasPtr->SaveAs(fileName.c_str(),option.c_str());
    }    

    //------------------------------------------------------------------------
    // saveHistogram():
    bool saveHistogram(string prefix="none", string ext="dat") {

      // Default file name is the name of display
      string fileName = displayNameSav;
      // If specific prefix given, append display name to that 
      if (prefix != "none") fileName = prefix+"-"+displayNameSav;	  
      fileName += "."+ext;
      FILE * pFile = fopen (fileName.c_str(), "w");
      if (pFile == NULL) {
	perror ("(VRdisplay::saveHistogram): Error opening file");
	return false;
      }
      else {
	// Number of bins to print
	TH1D* hisPtr = getTH1DPtr();
	int nBin = hisPtr->GetNbinsX();
	for (int iBin=1; iBin<=nBin; iBin++){
	  double dx  = hisPtr->GetBinWidth(iBin);
	  double xLo = hisPtr->GetBinLowEdge(iBin);
	  if ( abs(xLo/dx) < 1e-12 ) xLo = 0.;
	  double xHi = xLo + dx;
	  if ( abs(xHi/dx) < 1e-12 ) xHi = 0.;
	  double xC  = xLo + (xHi - xLo)/2.;
	  double y   = hisPtr->GetBinContent(iBin);
	  // Output standard 1-sigma uncertainties
	  double ey  = hisPtr->GetBinError(iBin);
	  // Systematic errors ?
	  double eyThHi(0.), eyThLo(0.);
	  if (graphs[0].hasOuter()) {
	    for (int iVar=0; iVar < graphs[0].nOuter(); ++iVar) {
	      eyThHi = max(eyThHi,getTH1DPtr(0,iVar)->GetBinContent(iBin) - y);
	      eyThLo = max(eyThLo,y - getTH1DPtr(0,iVar)->GetBinContent(iBin));
	    }
	  }

	  // If has systematic uncertainties
	  if (graphs[0].hasOuter()) {
	    fprintf(pFile,
		 "%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
		    xLo,xC,xHi,y,ey,ey,eyThHi,eyThLo); 
	  } else {
	    // Default: write out stat error only
	    fprintf(pFile,"%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
		    xLo,xC,xHi,y,ey,ey); 

	  }
	}

	// Write format as text comment
	if (graphs[0].hasOuter()) {
	  fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat          +dyTh          -dyTh\n");
	} 
	else {
	  fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat\n");
	}

	fprintf(pFile,"# Info:\nTIT = %s\n",hisPtr->GetTitle());
	if (hasLegend && legendSav != "") {
	  fprintf(pFile,"SRC = %s\n",legendSav.c_str());
	}
	if (hasMCnames && MCnamesSav != "") {
	  fprintf(pFile,"REF = %s\n",MCnamesSav.c_str());
	}
 	fprintf(pFile,"#\n# File printed by VinciaRoot\n");	
	fprintf(pFile,"# VinciaRoot is distributed as part of the VINCIA code\n"); 
	fprintf(pFile,"# http://projects.hepforge.org/vincia/\n"); 
	fprintf(pFile,"# http://arxiv.org/abs/arXiv:1102.2126\n");
	fclose(pFile);
	return true;
      }
      
    }

    //========================================================================
    // VRdisplay: Plot titles, legends, descriptions

    //------------------------------------------------------------------------
    // setTitle():
    // Set title
    void setTitle() {
      // Can only define title if histogram has been defined
      if (!hasAxes() || graphs.size() == 0) return;
      // Get current histogram title from histogram pointer
      TH1D* hisPtr = graphs[0].getTH1DPtr();
      string hisTitle = hisPtr->GetTitle();
      // Destroy pointer if title already exists (to avoid mem leak)
      if (hasTitle) delete titleTexPtr;
      // Get axis definitions from axis histogram
      int NBins = axesPtr->GetNbinsX();
      double xLo = axesPtr->GetBinLowEdge(1);
      double xHi = axesPtr->GetBinLowEdge(NBins+1);
      double xTit = xLo + (xHi-xLo)/2.;
      double yMin = axesPtr->GetMinimum();
      double yMax = axesPtr->GetMaximum();
      double dyTit = 0.07;
      double yTit = yMin + (1.-dyTit)*(yMax-yMin);
      if (gPad->GetLogy()) {
	double LnyC = log(yMin) + (1.-dyTit)*log(yMax/yMin);
	yTit = exp(LnyC);
      }      
      // Create new title tex pointer
      titleTexPtr = new TLatex(xTit,yTit,hisTitle.c_str());
      hasTitle    = true;
      titleTexPtr->SetTextFont(62);
      titleTexPtr->SetTextAlign(22);
      double titSize = (1.05+0.025*(nPad-1))*titleTexPtr->GetTextSize();
      titleTexPtr->SetTextSize(titSize);
      titleTexPtr->SetTextColor(1);
    }

    //------------------------------------------------------------------------
    // Set MC name title
    void setMCnames(string MCnamesIn) {
      if (!hasAxes() || graphs.size() == 0) return;
      // Destroy pointer if MCnames already exists (to avoid mem leak)
      if (hasMCnames) delete MCnamesPtr;
      // Get axis definitions from axis histogram
      int NBins = axesPtr->GetNbinsX();
      double xLo = axesPtr->GetBinLowEdge(1);
      double xHi = axesPtr->GetBinLowEdge(NBins+1);
      double xTit = xLo + (xHi-xLo)/2.;
      double yMin = axesPtr->GetMinimum();
      double yMax = axesPtr->GetMaximum();
      double dyTit = 0.085;
      double yTit = yMin + dyTit*(yMax-yMin);
      if (gPad->GetLogy()) {
	double LnyC = log(yMin) + dyTit*log(yMax/yMin);
	yTit = exp(LnyC);
      }      
      MCnamesPtr = new TLatex(xTit,yTit,MCnamesIn.c_str());
      MCnamesPtr->SetTextFont(42);
      MCnamesPtr->SetTextAlign(23);
      MCnamesPtr->SetTextColor(12);
      double namesSize = MCnamesPtr->GetTextSize();
      double scaleFac  = 1. + (nPad-1)/10.;
      MCnamesPtr->SetTextSize(0.7*scaleFac*namesSize);
      hasMCnames = true;
      MCnamesSav = MCnamesIn;
    }

    //------------------------------------------------------------------------
    // Set Data reference title
    void setDataRef(string dataRefIn) {
      if (!hasAxes() || graphs.size() == 0) return;
      // Destroy pointer if MCnames already exists (to avoid mem leak)
      if (hasDataRef) delete dataRefPtr;
      if (dataRefIn == "none") return;
      // Get axis definitions from axis histogram
      int NBins = axesPtr->GetNbinsX();
      double xLo = axesPtr->GetBinLowEdge(1);
      double xHi = axesPtr->GetBinLowEdge(NBins+1);
      double xTit = xLo + (xHi-xLo)/2.;
      double yMin = axesPtr->GetMinimum();
      double yMax = axesPtr->GetMaximum();
      double dyTit = 0.112;
      double yTit = yMin + dyTit*(yMax-yMin);
      if (gPad->GetLogy()) {
	double LnyC = log(yMin) + dyTit*log(yMax/yMin);
	yTit = exp(LnyC);
      }
      // Add "Data from ..." if missing
      if (dataRefIn.find("Data") == string::npos 
	  && dataRefIn.find("data") == string::npos) 
	dataRefIn = "Data from " + dataRefIn;      
      dataRefPtr = new TLatex(xTit,yTit,dataRefIn.c_str());
      dataRefPtr->SetTextFont(42);
      dataRefPtr->SetTextAlign(21);
      dataRefPtr->SetTextColor(12);
      double namesSize = dataRefPtr->GetTextSize();
      double scaleFac  = 1. + (nPad-1)/10.;
      dataRefPtr->SetTextSize(0.7*scaleFac*namesSize);
      hasDataRef = true;
    }

    //------------------------------------------------------------------------
    // Set Legend
    void setLegend(vector<string> legendIn) {
      if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0) return;
      // Destroy pointer if MCnames already exists (to avoid mem leak)
      if (hasLegend) delete legendPtr;
      // Find relUnc of axis
      int axisBins   = axesPtr->GetNbinsX();
      double axisMin = axesPtr->GetBinLowEdge(1);
      double axisCtr = 0.5*(axesPtr->GetBinLowEdge(axisBins+1) + axisMin);
      double xMax = axisMin;
      double yMax = -1e9;
      // Check if legend should be placed on left or right
      if (hasDataSav && dataPtr->GetN() >= 1) {
	int nPoints = dataPtr->GetN();
	double xTmp,yTmp;
	for (int iPoint = 0; iPoint < nPoints; iPoint++) {	  
	  dataPtr->GetPoint(iPoint,xTmp,yTmp);
	  if (yTmp > yMax) {
	    yMax = yTmp;
	    xMax = xTmp;
	  }
	}
      }
      if (graphs.size() >= 1) {
	for (int iGraph=0; iGraph < int(graphs.size()); ++iGraph) {
	  int nPoints = graphs[iGraph].GetN();
	  double xTmp,yTmp;
	  for (int iPoint = 0; iPoint < nPoints; iPoint++) {	  
	    graphs[0].GetPoint(iPoint,xTmp,yTmp);
	    if (yTmp > yMax) {
	      yMax = yTmp;
	      xMax = xTmp;
	    }
	  }
	}
      }
      // Set legend position Left or Right
      if ( xMax > axisCtr ) leftLegend = true;
      else leftLegend = false;

      // Legends use absolute coordinates (sucks)
      double xLeg = 0.56;
      if (leftLegend) xLeg = 0.18;
      double xLegEnd = xLeg + 0.33;
      double dy = 0.070;
      double yLegTop = 0.82;
      double yLegBot = yLegTop - dy*(legendIn.size()+hasDataSav);
      legendPtr = new TLegend(xLeg,yLegBot,xLegEnd,yLegTop,NULL,"rNDC");
      hasLegend = true;
      legendPtr->SetBorderSize(0);
      legendPtr->SetFillStyle(0);
      // Add data legend
      if (hasDataSav) {
	string datLeg = dataFilePtr->getSRC();
	if (datLeg == "" || datLeg == "none") datLeg = "Data";
	TLegendEntry *entry=legendPtr->AddEntry("gr",datLeg.c_str(),"p");
	entry->SetMarkerStyle(dataPtr->GetMarkerStyle());
	entry->SetMarkerSize(dataPtr->GetMarkerSize());
	entry->SetMarkerColor(dataPtr->GetMarkerColor());
	entry->SetTextAlign(12);
	entry->SetTextFont(42);
	entry->SetTextColor(1);	
      }
      // Add histogram legends
      for (int iLeg = 0; iLeg < min(int(graphs.size()), int(legendIn.size()));
           ++iLeg) {
	TLegendEntry *entry=legendPtr->AddEntry("gr",legendIn[iLeg].c_str(),"lP");
	entry->SetLineColor(graphs[iLeg].GetLineColor());
	entry->SetLineStyle(graphs[iLeg].GetLineStyle());
	entry->SetLineWidth(graphs[iLeg].GetLineWidth());
	entry->SetMarkerStyle(graphs[iLeg].GetMarkerStyle());
	entry->SetMarkerSize(graphs[iLeg].GetMarkerSize());
	entry->SetMarkerColor(graphs[iLeg].GetMarkerColor());
	entry->SetTextAlign(12);
	entry->SetTextFont(42);
	entry->SetTextColor(1);
      }
      // Save legend of main (=first) histogram for use by saveHistogram().      
      legendSav = (legendIn.size() > 0) ? legendIn[0] : "";
    }
    
  private:

    //========================================================================
    // VRdisplay: private data members and functions
    
    // Canvas
    string displayNameSav;
    TCanvas* displayCanvasPtr;

    // Internal variables
    bool logY, normalize;
    bool showRatioPad, showRelUncPads;
    int verbose, optStat;
    double sigmaStat, normFac;

    // Input containers
    VRfile* dataFilePtr;
    vector<VRfile*> hisFilePtrs;
    
    // Global settings
    VRsettings* settingsPtr;

    // PADS
    int nPad;
    bool isInitPad;

    // PLOT

    // Use histogram to define axes
    TH1D* axesPtr;
    TH1D* axesNormPtr;
    TH1D* axesVarPtr;
    bool hasAxesSav;

    // MC Graphs 
    vector<VRgraph> graphs;
    vector<VRgraphNorm> graphsNorm;
    // RelUnc pad
    vector<VRgraphVar> graphsVar;

    // Data graph
    VRgraphData* dataPtr;
    VRgraphNorm* dataNormPtr;
    bool hasDataSav;

    // Plot title 
    TLatex* titleTexPtr;
    bool hasTitle;
    
    // LEGEND

    // Main Legend 
    TLegend* legendPtr;
    bool hasLegend;
    bool leftLegend;
    string legendSav;

    // MC names and versions
    TLatex* MCnamesPtr;
    bool hasMCnames;
    string MCnamesSav;
    
    // Data reference
    TLatex* dataRefPtr;
    bool hasDataRef;

    // VinciaRoot Signature (on lower right-hand edge of plots)
    TLatex* signaturePtr;
    bool hasSignature;
    string signature;

  };

  //***************************************************************
  // class VinciaRoot: Main VinciaRoot Class
  // Note: is completely independent of VINCIA and can be used 
  // with any executable using ROOT histograms (specifically TH1F)
  //***************************************************************

  class VinciaRoot {

  public:

    // Constructor
    VinciaRoot() {

      // Switch on storage of stat uncertainties 
      TH1::SetDefaultSumw2();

    }
    
    // Initialize ROOT app
    void initApp() {
      cout<<"(VinciaRoot): initializing display app"<<endl;
      gROOT->Reset();
      gROOT->SetStyle("Plain");	
      TApplication *vcApp = new TApplication("vcApp",0,0);
      vcApp->InitializeGraphics();
      settings.isInitApp = true;
    }      

    // Check if a display exists
    bool exists(string displayName) {
      if (displays.find(displayName) == displays.end()) return false;
      else return true;
    }

    // Initialize a display
    void createDisplay(string displayName, int xSizIn=300, int ySizIn=400, 
		     int xOffsetIn=-1,int yOffsetIn=-1) {

      // Check if Root display app has been initialized
      if ( !settings.isInitApp ) initApp(); 

      // Check if display with that name already exists
      if ( exists(displayName) ) {
	cout<<"(VinciaRoot): warning! a display named "<<displayName
	    <<" already exists"<<endl;
	displays.erase(displayName);
      }

      // Cascade windows from top right corner if no offsets specified
      if (xOffsetIn == -1) xOffsetIn = 16*displays.size();
      if (yOffsetIn == -1) yOffsetIn = 16*displays.size();

      // Create and initialize display 
      displays[displayName] = new VRdisplay(displayName,xSizIn,ySizIn,
					    xOffsetIn,yOffsetIn);
      displays[displayName]->init(&settings);
 
    }

    // Create a new Display from TH1D*
    void createDisplay(string  displayName, TH1D* HisPtrIn, 
		       int xSizIn=300, int ySizIn=400, 
		       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->addTH1DPtr(HisPtrIn);
    }
    
    // Create a new Display from vector<TH1D*>
    void createDisplay(string displayName, vector<TH1D*> HisPtrIn, 
		       int xSizIn=300, int ySizIn=400, 
		       int xOffsetIn=-1,int yOffsetIn=-1) {
      // Initialize display and associate 
      createDisplay(displayName, xSizIn, ySizIn, xOffsetIn, yOffsetIn);
      displays[displayName]->addTH1DPtr(HisPtrIn);
    }

    // (Re)set global verbose level
    void setVerbose(int verboseIn) {
      settings.verbose = verboseIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) it->second->setVerbose(verboseIn);
    }
    
    // (Re)set global optstat
    void setOptStat(int optStatIn) {
      settings.optStat = optStatIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) 
	it->second->setOptStat(optStatIn);
    }

    // Set global optstat
    void setSigmaMC(double sigmaStatIn) {
      settings.sigmaStat = sigmaStatIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) 
	it->second->setSigmaMC(sigmaStatIn);
    }

    // Set normalize on/off
    void setNormalize(string displayNameIn, bool normalizeIn = true, 
		      double normFacIn = 1.0) {
      if (exists(displayNameIn)) displays[displayNameIn]->setNormalize(normalizeIn,
								       normFacIn);
      else cout<<"(VinciaRoot): setNormalize no display named "<<displayNameIn<<endl;
    }

    // Set logarithmic Y axis
    void setLogY(string displayNameIn, bool logY=true) {
      if (exists(displayNameIn)) displays[displayNameIn]->setLogY(logY);
      else cout<<"(VinciaRoot): setLogY no display named "<<displayNameIn<<endl;
    }
    void setLogY(bool logY=true) {
      settings.logY = logY;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) 
	it->second->setLogY(logY);
    }
    // Set linear Y axis
    void setLinY(string displayNameIn, bool linY=true) { 
      if (exists(displayNameIn)) displays[displayNameIn]->setLogY(!linY); 
      else cout<<"(VinciaRoot): setLinY no display named "<<displayNameIn<<endl;
    }
    void setLinY(bool linY=true) { setLogY(!linY); }

    // Set MC names
    void setMCnames(string displayNameIn, string MCnamesIn) {
      if (exists(displayNameIn)) displays[displayNameIn]->setMCnames(MCnamesIn);
      else cout<<"(VinciaRoot): setMCnames no display named "<<displayNameIn
	       <<endl;
    }
    void setMCnames(string MCnamesIn) {
      settings.MCnames = MCnamesIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) 
	it->second->setMCnames(MCnamesIn);
    }

    // Set legend
    void setLegend(string displayNameIn, vector<string> legendIn) {
      if (exists(displayNameIn)) displays[displayNameIn]->setLegend(legendIn);
      else cout<<"(VinciaRoot): setLegend no display named "<<displayNameIn
	       <<endl;
    }
    void setLegend(vector<string> legendIn) {
      settings.legend = legendIn;
      // Propagate change to all displays
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) 
	it->second->setLegend(legendIn);
    }

    // Wrapper for passing single legend as string
    void setLegend(string displayNameIn, string legendIn) {
      vector<string> legendVec;
      legendVec.push_back(legendIn);
      setLegend(displayNameIn,legendVec);
    }
    void setLegend(string legendIn) {
      vector<string> legendVec;
      legendVec.push_back(legendIn);
      setLegend(legendVec);
    }

    // Functions to give access to member functions of individual displays
    void addTH1DPtr(string displayName, TH1D* hisPtrIn) {
      if ( !exists(displayName) ) createDisplay(displayName, hisPtrIn);
      else displays[displayName]->addTH1DPtr(hisPtrIn);
    }
    void addTH1DPtr(string displayName, vector<TH1D*> hisPtrIn) {
      if ( !exists(displayName) ) createDisplay(displayName, hisPtrIn);
      else displays[displayName]->addTH1DPtr(hisPtrIn);
    }    
    void setDataFile(string displayName, string fileIn, bool hasSysErrors=true) {
      if (hasSysErrors) setDataFile(displayName,fileIn,10,20,30,40,51,52,61,62);
      else setDataFile(displayName,fileIn,10,20,30,40,51,52);
    }
    void setDataFile(string displayName, string fileIn, 
		     int i1, int i2, int i3=-1,
		     int i4=-1, int i5=-1, int i6=-1, int i7=-1, int i8=-1, 
		     int i9=-1, int i10=-1, int i11=-1, int i12=-1) {
      if ( !exists(displayName) && settings.verbose >= 1) {
	cout<<"(VinciaRoot): ERROR! display "<<displayName
	    <<" does not exist. Attempt to add data file failed."<<endl;
	return;
      }
      displays[displayName]->setDataFile(fileIn, i1, i2, i3, i4, i5, i6, i7,
					 i8, i9, i10, i11, i12);
    }    
    void addHisFile(string displayName, string fileIn,  
		    bool hasSysErrors=true) {
      if (hasSysErrors) addHisFile(displayName,fileIn,10,20,30,40,51,52,61,62);
      else addHisFile(displayName,fileIn,10,20,30,40,51,52);
    }
    void addHisFile(string displayName,string fileIn, int i1, int i2, int i3=-1,
		     int i4=-1, int i5=-1, int i6=-1, int i7=-1, int i8=-1, 
		     int i9=-1, int i10=-1, int i11=-1, int i12=-1) {
      if ( !exists(displayName) && settings.verbose >= 1) {
	cout<<"(VinciaRoot): ERROR! display "<<displayName
	    <<" does not exist. Attempt to add histogram file failed."<<endl;
	return;
      }
      displays[displayName]->addHisFile(fileIn, i1, i2, i3, i4, i5, i6, i7,
					 i8, i9, i10, i11, i12);
    }    
    
    // Save Main histograms (with uncertainties)
    void saveHistograms(string baseName, string fileType) {
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) it->second->saveHistogram(baseName, 
								 fileType);
    }

    // Update display(s): default is update all
    void update(string displayName="all") {
      if (displayName == "all") {
        for (map<string,VRdisplay*>::iterator it=displays.begin(); 
             it != displays.end(); ++it) it->second->update();	
      } else if ( exists(displayName) ) {
        displays[displayName]->update();	
      }
    }
    // Legacy functions for backwards compatibility
    void updateDisplay(string displayName="all") { update(displayName); }
    void updateDisplays() { update(); }

    // Save Display
    void saveDisplay(string displayNameIn, 
		     string baseName, string fileType="png") {
      if (exists(displayNameIn)) displays[displayNameIn]->saveDisplay(baseName,
								    fileType);
      else cout<<"(VinciaRoot): saveDisplay no display named "<<displayNameIn
	       <<endl;
    }
    
    void saveDisplays(string baseName, string fileType="png") {
      for (map<string,VRdisplay*>::iterator it=displays.begin(); 
	   it != displays.end(); ++it) it->second->saveDisplay(baseName, 
							       fileType);
    }

    // Return pointer to an entire display (if desired)
    VRdisplay* getDisplay(string displayName) { 
      if (exists(displayName)) return displays[displayName];
      else {
	cout<<"(VinciaRoot): ERROR! attempt to access non-existing display "
	    <<displayName<<" - retuning NULL"<<endl;
	return NULL;
      }
    }

  private:

    // vector of displays
    map<string,VRdisplay*> displays;

    // Global settings
    VRsettings settings;

  };

  //***************************************************************
  // class UncertaintyBand: 
  // Extremely simple auxiliary class for storing names and indices 
  // of uncertainty bands
  //***************************************************************

  class UncertaintyBand {
    
  public: 
    
    // Constructor
    UncertaintyBand(string nameIn=" ") {
      name = nameIn;
      indices.resize(0);
    };
    
    // Name
    string name;
    // Indices
    vector<int> indices;
    
  };
 
}

#endif
