// VinciaTools.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2012 P.Z.Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's Toolbox.
  
// Classes:
// VinClu : VINCIA's clustering maps
// Rambo  : Flat phase-space generator

#ifndef VINCIA_VinciaTools_H
#define VINCIA_VinciaTools_H

#include "Basics.h"             // Vec4
#include "VinciaCommon.h"       // Vec4toD4, m2(...)

using namespace Pythia8;

namespace Vincia {

  //**************************************************************************

  // VINCIA Clustering Map

  class VinClu {

  public:
    
    // Constructor.
    VinClu() {isInitPtr=false; isInit=false;}

    // Destructor.
    virtual ~VinClu() {}

    // Initialize pointers
    void initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn, 
		 VinciaCommon* vinComPtrIn) {
      rndmPtr     = rndmPtrIn; 
      settingsPtr = settingsPtrIn; 
      vinComPtr   = vinComPtrIn;
      isInitPtr   = true;
    }

    // Intialize VinClu (currently just to set verbose level)
    bool init() {
      verbose = settingsPtr->mode("Vincia:verbose");
      isInit = true;
      return true;
    }

    // Function to cluster a set of partons
    // (basically vc3to2 overlaid with vectors of parton IDs from which
    //  the appropriate clustering is automatically determined and the 
    //  correct clustering map retrieved from the corresponding antenna)
    bool getClu(vector<int> idIn, vector<Vec4> pIn,int ia, int ir, int ib, 
		vector<int>& idClu, vector<Vec4>& pClu, 
		int& antIndx, bool& swapAB);

    // Main Vincia clustering map
    // MR 12.7.2010: changed to return bool
    bool vc3to2(int iMap,vector<Vec4>,int,int,int,vector<Vec4>&);
	
    // MR 7.7.2010: massive generalization of the clustering map, takes
    // the two parent masses as additional arguments
    // MR 12.7.2010: changed to return bool
    bool vc3to2massive(int iMap, vector<Vec4>, int, int, int, double,
		       double, vector<Vec4>&);
    
    //******

    // Individual maps, accessed by the main algorithm
    // Imap = 1 & 2 so far accessed by VC3TO2 fortran routine
	
    // MR 11.7.2010: more general version accepting kineMapType
    // as the first argument, changed to return a boolean
    bool map3to2(int kineMapType, Vec4 pi, Vec4 pj, Vec4 pk,
		 Vec4& pOutI, Vec4& pOutK);
    
    // MR 7.7.2010: massive generalization of 3->2 clustering, takes
    // map type as the first argument
    bool map3to2massive(int kineMapType, Vec4 pi, Vec4 pj, Vec4 pk,
			double mI,double mK, Vec4& pOutI,Vec4& pOutK);

    // Check if 2-particle system is on-shell and rescale if not
    bool onShellCM(Vec4& p1, Vec4& p2, double m1, double m2) {
      double s1     = pow2(m1);
      double s2     = pow2(m2);
      double s12    = Vec4(p1+p2).m2Calc();
      double s1Calc = p1.m2Calc();
      double s2Calc = p2.m2Calc();
      if (abs(s1Calc-s1)/s12 > TINY || abs(s2Calc-s2)/s12 > TINY ) {
	if (verbose >= 3) 
	  printOut("VinClu::onShellCM","forcing particles on mass shell");
	RotBstMatrix M;
	M.fromCMframe(p1,p2);
	// Define massive on-shell momenta
	double E1 = (s12 + s1 - s2)/(2*sqrt(s12));
	double E2 = (s12 - s1 + s2)/(2*sqrt(s12));
	double pz = pow2(E1)-s1;
	Vec4 p1new = Vec4(0.0,0.0,-pz,E1);
	Vec4 p2new = Vec4(0.0,0.0,pz,E2);
	p1new.rotbst(M);
	p2new.rotbst(M);
	double s1Test = p1new.m2Calc();
	double s2Test = p2new.m2Calc();
	if (verbose >= 3) {
	  cout<<" p1   : "<<p1<<" p1new: "<<p1new;
	  cout<<" p2   : "<<p1<<" p2new: "<<p1new;
	}
	// If this got them closer to mass shell, replace momenta
	if (abs(s1Test-s1)/s12 <= abs(s1Calc-s1)/s12
	    && abs(s2Test-s2)/s12 <= abs(s2Calc-s2)/s12) {
	  p1 = p1new;
	  p2 = p2new;
	}
	return false;
      }
      else return true;
    } 

    // Get/Set parameters
    int getVerbose() {return verbose;}
    void setVerbose(int verboseIn) {verbose=verboseIn;}

  private :
  
    // Is initialized?
    bool isInitPtr, isInit;

    // Pointers to VINCIA and PYTHIA 8 objects
    Rndm*  rndmPtr;
    Settings* settingsPtr;
    VinciaCommon* vinComPtr;

    // Verbose level    
    int verbose;

  };

  //**************************************************************************

  // Rambo flat phase-space generator

  class Rambo {
   
  public:
    
    // Deafult Constructor.
    Rambo() {isInitPtr=false;}
    
    // Initializing Constructor.
    Rambo(Rndm* rndmPtrIn) { initPtr(rndmPtrIn); }

    // Destructor.
    virtual ~Rambo() {}
    
    // Initialize pointers
    void initPtr(Rndm* rndmPtrIn) {rndmPtr = rndmPtrIn; isInitPtr=true;}
    
    // Rambo phase space generator
    // p[0][i] is the energy of the i'th parton
    double genPoint(double S,int N,double p[4][8]);
    // MR 6.2.2011: massive generalisation, weights NOT 1 anymore
    // literal implementation of original RAMBO paper by Ellis, 
    // Kleiss and Stirling
    double genPoint(double S,int N,double m[8],double p[4][8]);
    // MR 15.3.2011: implement a function which returns a sequence 
    // of massless vectors in which the n1 and n2 become 
    // collinear
    vector< vector<Vec4> > collinearSequence(
      double S, int nParticles, int nSteps, int n1, int n2, 
      double rescalingFactor, double rescalingFactorTolerance = 1e-3 );
    // MR 28.2011: re-implement the RAMBO algorithm according to 
    // the original paper again as a check
    double genPointPrime(
      double ecm, int n, vector<double> masses, vector<Vec4>& retVec);
    
  private:
    
    // Is initialized?
    bool isInitPtr;

    // Pointer to the random number generator.
    Rndm*  rndmPtr;
    
  };

}

#endif
