// VinciaMatching.cc is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2012 P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the Matching class. 

#include "VinciaMatching.h"

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  // init() : 
  // Initialize Matching interface
  //***********************************************************************

  bool Matching::init(int nVariationsIn) {

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("Matching::init","Cannot initialize, pointers not set.");
      return false;
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Set number of uncertainty variations (at least one, for USR)
    nVariations = max(1,nVariationsIn);

    // Initialize Settings
    verbose          = settingsPtr->mode("Vincia:verbose");
    fullColor        = settingsPtr->flag("VinciaMatching:fullColor");
    uncertaintyBands = settingsPtr->flag("Vincia:uncertaintyBands");
    // Initialize Shower Settings (needed to compute matching coefficients)
    evolutionMode    = settingsPtr->mode("Vincia:evolutionMode");
    sectorOrdering   = settingsPtr->flag("VinciaSector");
    showerNLC        = settingsPtr->mode("Vincia:NLC");

    // Check and initialize MG interface 
    isMGOn = settingsPtr->flag("Vincia:MGInterface");
    if (isMGOn) isMGOn = mgInterfacePtr->init();

    // Cannot match if MG missing
    isInit = isInit && isMGOn;

    return isInit;
  }
	
  //***********************************************************************
  // Main functions for VINCIA matching
  //***********************************************************************

  // *********************************************************************
  // isMatched() : return true if any matching exists at this order
  // *********************************************************************

  bool Matching::isMatched(int idRes, int nEmit, int nSplit, bool tryMG) {
    
    // Return value false by default
    bool gotit = false;

    // -------------------------------------------------------------------
    // Matching: with MG interface
    // -------------------------------------------------------------------
    if (tryMG) {

      // Spin-1 Charge-0 Color-Singlet (Z0) Decays
      if (idRes == 23) {
	if (nEmit+nSplit <= 4) gotit = true;
      } 
      
      // Spin-1 Charge-1 Color-Singlet (W) Decays
      // (currently using same as Charge-0)
      else if (idRes == 24) {
	if (nEmit+nSplit <= 4) gotit = true;
      }
      
    }

    return gotit;

  } 
  
  // *********************************************************************
  // getPME() : Return a matching coefficient
  // INPUT: 
  //      nIn : 1 (for decay) or 2 (for 2->2 process)
  //      id  : vector containing PDG ID codes, 
  //           first incoming parton(s), then outgoing ones
  //      p   : same as is, for parton 4-vectors
  //   Note   : the parton list is assumed to be color-ordered
  // OUTPUT: 
  //      PME : vector whose [0] component is the matching coefficient
  //            for the current shower settings, with the additional ones
  //            corresponding to uncertainty variations (if switched on).
  //              = |M_n|^2 / (Sum_i A^phys_i |M_n-1|^2)
  //      PME[0]   : reweighting factor with USR antennae
  //      PME[1]   : reweighting factor with DEF antennae 
  //      PME[2,3] : no difference at tree level (muR-variation)
  //      PME[4]   : reweighting factor with MAX antennae
  //      PME[5]   : reweighting factor with MIN antennae
  //      PME[6,7] : no difference at tree level (NLO K-factor variation)
  //      PME[8]   : reweighting factor with stronger ordering (Pimp^2)
  //      PME[9]   : reweighting factor with mD ordering
  //      PME[10]  : reweighting factor with all emission color factors = CA 
  //      PME[11]  : reweighting factor with all emission color factors = CF
  // *********************************************************************

  bool Matching::getPME(int nIn, vector<int> id, vector<Vec4> p, 
			vector<double>& PME) {

    // Resize PME.
    PME.resize(0);
    bool done = false;

    // Matching using MadGraph
    if (! isMGOn) return false;

    // Get color-ordered M2 (= LC + ...) for this n-parton state
    double Mn2     = mgInterfacePtr->ME2(nIn, id, p);
    
    // mgInterface returns -1 if not existing. 
    if (abs(Mn2+1) < TINY) return false;

    // Initialize (n-1)->n vector of approximations
    vector<double> Mn2Appr(0);
    for (int iVar=0; iVar<nVariations; iVar++) {
      Mn2Appr.push_back(0.);
    }

    // SECTOR showers: find appropriate sector    
    // Loop over possible sectors, start with first one
    int irMin    = 2;
    if (sectorOrdering) {
      double q2Min = RPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],1);
      // Juan to insert sqrt requirement here for quarks ...
      // Loop over
      for (int ir=3; ir<int(p.size())-1; ++ir) {
	double q2 = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],1);
	if (q2 < q2Min) {
	  q2Min = q2;
	  irMin = ir;
	}
      }    
    }    

    // Loop over possible emitter antennae
    for (int iClu=0; iClu<int(id.size())-3; iClu++) {
      
      // Partons to be clustered
      int ir    = 2+iClu;

      // SECTOR showers: skip all clusterings except correct sector
      if (sectorOrdering && ir != irMin) continue;

      // Compute basic quantities
      double ma = ( vinComPtr->isMassive(id[ir-1]) ? p[ir-1].mCalc() : 0.0 ) ; 
      double mr = ( vinComPtr->isMassive(id[ir])   ? p[ir].mCalc() : 0.0 );
      double mb = ( vinComPtr->isMassive(id[ir+1]) ? p[ir+1].mCalc() : 0.0 );
      double sP  = m2(p[ir-1]+p[ir]+p[ir+1]);
      double mP  = sqrt(sP);
      double mMua = ma / mP;
      double mMur = mr / mP; 
      double mMub = mb / mP;
      
      // Reconstruct previous state
      vector<int> idClu;
      vector<Vec4> pClu;
      int antIndx=-1;
      bool swapAB=false;
      vinCluPtr->getClu(id,p,ir-1,ir,ir+1,idClu,pClu,antIndx,swapAB);
      
      // Invariant mass of parent antenna and daughter invariants
      double yar = (m2(p[ir-1],p[ir])-pow2(ma)-pow2(mr))/sP;
      double yrb = (m2(p[ir],p[ir+1])-pow2(mr)-pow2(mb))/sP;
      if (swapAB) {
	double ytmp = yar;
	yar = yrb;
	yrb = ytmp;
        // MR 17.1.2011: have to swap masses as well
        double mMutmp = mMua;
        mMua = mMub;
        mMub = mMutmp;
      }
      
      // Check for vanishing invariants
      if (yar < TINY || yrb < TINY) {
	printOut("VinciaMatching::getPME","Warning: small invariant(s): y = "
		 +num2str(yar)+","+num2str(yrb)
		 +" sParent ="+num2str(sP));
	if (verbose >= 3) {
	  cout<<"Momentum Listing:"<<endl;
	  cout<<p[ir-1]<<p[ir]<<p[ir+1];
	  cout<<"Invariants: "<<m2(p[ir-1]+p[ir])<<" "<<m2(p[ir]+p[ir+1])<<endl;
	}
      }
      
      // Compute Pimp factor 
      //  : 1 for evolutionMode <= 1
      //  : Q2E/(Q2Ehat+Q2E) for type 2 ordering
      //  : pT2E/(pT2Ehat+pT2E) for type 3 ordering
      double Pimp    = 1.0;
      double PimpPT  = 1.0;
      double PimpMD  = 1.0;
      double PimpPT2 = 1.0;
      // Only apply if at least 3 partons in final state
      if ( evolutionMode >= 2 && int(pClu.size()-2) > nIn) { 
	// Decide whether to use pT or QE for unordering suppression
	int uoType = 1;
	if (evolutionMode == 2) {
	  // Note: if using different evolution types for different 
	  // branchings, this is the one for the most recent branching
	  uoType = antennaSets[0]->getAnt(antIndx)->getEvolutionType();
	}
	// Compute suppression factor
	double pT2Hat = RPtr->Q2E(pClu[1],pClu[2],pClu[3],1);
	double mD2Hat = RPtr->Q2E(pClu[1],pClu[2],pClu[3],2);
	for (int jClu=1; jClu<int(pClu.size())-3; jClu++) {
	  int jaHat=jClu+1;
	  int jrHat=jClu+2;
	  int jbHat=jClu+3;
	  pT2Hat = min(pT2Hat,
		       RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],1));
	  mD2Hat = min(mD2Hat,
		       RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],2));
	}	  
	double pT2New = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],1);
	double mD2New = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],2);
	PimpPT  = pT2Hat/(pT2Hat+pT2New);
	PimpPT2 = pow2(pT2Hat)/(pow2(pT2Hat)+pow2(pT2New));
	PimpMD  = mD2Hat/(mD2Hat+mD2New);
	Pimp    = PimpPT;
	if (uoType == 2) Pimp = PimpMD;
	if (verbose >= 8) 
	  printOut("Matching::getPME","Pimp = "+num2str(Pimp));
      }
      
      // Compute ARIADNE factor (for g->qqbar splittings)
      double Pari = 1.0;
      if (antIndx == iqgSplit || antIndx == iggSplit) {
	double sN;
	if (!swapAB) {
	  // Only apply if neighbor is in final state
	  if (ir+1 < int(pClu.size()) ) { 
	    sN = m2(pClu[ir],pClu[ir+1]);
	    // Only apply if neighbor is a quark
	    if (idClu[ir+1] != 21) 
	      Pari = 2.0*sN/(sP + sN);
	  }
	} else { 
	  // Only apply if neighbor is in final state
	  if (ir-2 >= nIn) {
	    sN = m2(pClu[ir-2],pClu[ir-1]);	
	    // Only apply if neighbor is a quark
	    if (idClu[ir-2] != 21)  
	      Pari = 2.0*sN/(sP + sN);
	  }
	}
	
	// Verbose output
	if (verbose >= 8) 
	  printOut("Matching::getPME","Pari = "+num2str(Pari));
	
      }
      
      // Total prefactor
      double comFac = Pimp * Pari;
      
      // Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
      double Mp2 = mgInterfacePtr->ME2(nIn, idClu, pClu);
      if (Mp2 <= 0.0) {
	string idString=num2str(idClu[0])+" -> ";
	for (unsigned int i=1;i<idClu.size();i++) idString+=num2str(idClu[i]);
	printOut("VinciaMatching::getPME","Unphysical momenta or unknown ME for "+idString);
	// Print momentum configuration
	if (verbose >= 3) {
	  for (unsigned int i=0;i<pClu.size();i++) {
	    cout<< "    i = "<<i<<" px = "<<num2str(pClu[i].px())<<" py = "<<num2str(pClu[i].py())<<" pz = "<<num2str(pClu[i].pz())<<" e = "<<num2str(pClu[i].e())<<" m = "<<pClu[i].mCalc()<<endl;
	  }
	}
      }

      // TMP!
      if (verbose >= 9) 
	printOut("VinciaMatching::getPME","PME Mp2 = "+num2str(Mp2));
	      
      // Check if return was -1 ?
      if (abs(Mp2+1.0) < TINY) return false; 
      
      // Loop over different antenna sets (for uncertainty evaluation)
      // Note: we do not make modifications for strong ordering here, since
      // we assume unordered generation up to the matched orders, and hence
      // the matching corrections should always include unordering. 
      // We also do not make variations in alphaS here, since all powers
      // of the coupling are factored out of the matrix element correction.
      for (int iVar=0; iVar<nVariations; iVar++) {
	
	// Compute physical antenna for this set for this (n-1)-parton state
	int iSet = 0;
	if (iVar >= 1) iSet = 1;
	if (iVar == 4) iSet = 2;
	if (iVar == 5) iSet = 3;
	Antenna* thisAnt = antennaSets[iSet]->getAnt(antIndx);
	double antPhys = thisAnt->aBarY(yar,yrb,mMua,mMur,mMub)/sP;
	double chargeFactor = thisAnt->getChargeFactor();

	// SECTOR showers
	if (sectorOrdering) {
	  // Gluon emission: add IJ and JK collinear terms when applicable
	  if (id[ir] == 21) {
	    if (id[ir-1] == 21) antPhys += 2./yar/(1-yrb)
				  - 2./yar + yrb/yar - pow2(yrb)/yar ;
	    if (id[ir+1] == 21) antPhys += 2./yrb/(1-yar) 
				  - 2./yrb + yar/yrb - pow2(yar)/yrb;
	  }
	  // Gluon splitting just gets overall factor 2
	  else if (antIndx == iqgSplit || antIndx == iggSplit) antPhys *= 2.0;
	}
	
	  // Uncertainty variations
	if (iVar == 8) antPhys *= PimpPT2/Pimp;
	if (iVar == 9) antPhys *= PimpMD/Pimp;
	// Subleading-Color variations
	// (If doing full-color matching and NLC showering, don't make
	//  any variation here.)
	// If doing LC showering, the variation is canceled by var in shower
	// If doing LC matching, this helps estimate NLC uncertainty
	if (!showerNLC && fullColor) {
	  if (iVar == 10 && id[ir] == 21) chargeFactor = CA;
	  if (iVar == 11 && id[ir] == 21) chargeFactor = CF;
	}
	antPhys *= chargeFactor;
	antPhys *= comFac;
	// Add to Approximated Mn2 for this set
	Mn2Appr[iVar] += Mp2 * antPhys;
	
      }
      
    }
    
    // Store ME reweight factors
    for (int iVar=0; iVar<nVariations; iVar++) {
      if (iVar >= 1 && ! uncertaintyBands) continue;
      PME.push_back(Mn2/Mn2Appr[iVar]);
    }
  
    // Diagnostics if shower VERY bad approximation
    if (PME[0] > 1000.0 || (verbose >= 3 && PME[0] > 100.0 )) {
      printOut("VinciaMatching::getPME","Warning: large correction at order n ="
	       +num2str(int(p.size()-1),2)
	       +" unregulated PME[0] = "+num2str(PME[0])
	       +" ME"+num2str(int(p.size()-1),1)+" = "+num2str(Mn2)
	       +" Sum(PS*ME"+num2str(int(p.size()-2),1)+") = "+num2str(Mn2Appr[0]));
    }

    // Store Old ME (future possibility for speedup)
    // oldME2 = Mn2;
    
    // Done
    done = true;
    
    // For the time being, no non-MadGraph matching implemented
    return done;

  }

  //*******

  // get2to4():
  // Return 2->4 reweighting factors 
  //   id  : id codes for the 4 partons
  //   p   : 4-momenta for the 4 partons
  // Return reweighting factors in PME
  //   P2to4[0] : reweighting factor with USR antennae
  //   P2to4[n] : reweighting factor with variations, see getPME()
  // = |A_4|^2 / (Sum_i A^phys_i A^phys_i_n-1)

  bool Matching::getP2to4(vector<int> id, vector<Vec4> p, 
			vector<double>& P2to4) {

    // Resize P2to4.
    P2to4.resize(0);
    bool done = false;

    // Check if at least 4 partons in final state
    if (id.size() <= 5) return false;

    // Initialize (n-1)->n vector of approximations
    vector<double> Mn2Appr(0);
    for (int iVar=0; iVar<nVariations; iVar++) {
      Mn2Appr.push_back(0.);
    }
    
    // SYSTEMATIC 2->4 MATCHING
    // Requires that collinear singularities correctly
    // partitioned in 2->4 functions

    // Loop over possible emitter antennae
    for (int iClu=0; iClu<int(id.size())-3; iClu++) {
      
      // Clustering:
      //cout<<" P2to4 Clustering "<<iClu<<"/"<<int(int(id.size())-4)<<endl;
      
      // Compute new ordering scale
      int ir  = 2+iClu;
      int ida = id[ir-1];
      int idr = id[ir];
      int idb = id[ir+1];
      
      //cout <<"  arb = "<<ir-1<<" "<<ir<<" "<<ir+1<<" QNew = "<<sqrt(Q2New)<<endl;
      
      // B40A: (q qbar) -> (q g) qbar -> q q'bar q' qbar
      // B40B: (q qbar) -> q (g qbar) -> q q'bar q' qbar
      
      // E40aA: (q g) -> (q qbar' q') -> q g qbar' q'
      // (match after gluon emission inside qqbar antenna)
      // E40aB: (q g) -> q (g g) -> q g qbar' q'
      // (match after gluon splitting inside gg antenna, outside)
      
      // E40bA: (q g) -> (q g) g -> q q'bar q' g
      // (match after gluon splitting inside qg antenna
      // E40bB: (q g) -> q (g g) -> q q'bar q' g
      // (match after gluon splitting inside gg antenna, inside)
      
      // H40A: (g g) -> qbar (q g) -> qbar q qbar' q'
      // (match after gluon splitting inside qg antenna
      // H40B: (g g) -> (g qbar) q -> qbar q qbar' q'
      // (match after gluon splitting inside gqbar antenna
      
      // G40A: (g g) -> (g g) g -> g qbar q g
      // (match after gluon splitting inside qg antenna
      // G40B: (g g) -> g (g g) -> g qbar q g
      // (match after gluon splitting inside gqbar antenna
      
      // Reconstruct previous state
      vector<int> idClu(id);
      vector<Vec4> pClu;
      int idaHat  = ida;
      int idbHat  = idb;
      int antIndx = -1;
      bool swapAB = false;
      if (idr == 21) {
	// To recluster gluon, just collapse parent ids
	idClu.erase(idClu.begin()+ir);
	if (ida == 21 && idb == 21) antIndx = iggEmit;
	else if (idb == 21) antIndx = iqgEmit;
	else if (ida == 21) {antIndx = iqgEmit; swapAB=true;}
	else antIndx = iqqEmit;
      } else if (idr < 0) {
	// To recluster antiquark, collapse with downstream quark 
	idClu[ir+1]=21;
	idClu.erase(idClu.begin()+ir);
	idbHat = 21;
	if (idaHat == 21) antIndx = iggSplit;
	else antIndx = iqgSplit;
      } else {
	// To recluster quark, collapse with upstream antiquark
	idClu[ir-1]=21;
	idClu.erase(idClu.begin()+ir);
	idaHat = 21;
	if (idbHat == 21) {antIndx = iggSplit; swapAB=true;}
	else {antIndx = iqgSplit; swapAB=true;}
      }
      if (verbose >= 9) {
	cout << "P2to4 Clustering "<<iClu<<" : ";
	for (int i=0;i<int(idClu.size());i++) cout << " " << idClu[i];
	if (antIndx == iqgEmit) cout << (swapAB ? " gqbarEmit" : " qgEmit");
	cout<<" Ant = "<<antennaSets[0]->getAnt(antIndx)->humanName();
	cout<<endl;
      }
      // Reconstruct clustered momenta
      int iMap = antennaSets[0]->getAnt(antIndx)->getKineMapType();
      vinCluPtr->vc3to2(iMap,p,ir-1,ir,ir+1,pClu);
      if (verbose >= 9) {
	for (int i=0;i<int(idClu.size());i++) {
	  cout << "   p" << i<<" "<<pClu[i].px()<<" "<<pClu[i].py()
	       <<" "<<pClu[i].pz()<<" "<<pClu[i].e()<<endl;
	}
      }
      
      // Invariant mass of parent antenna and daughter invariants
      double sP  = m2(pClu[ir-1],pClu[ir]);
      double yar = m2(p[ir-1],p[ir])/sP;
      double yrb = m2(p[ir],p[ir+1])/sP;
      if (swapAB) {
	double ytmp = yar;
	yar = yrb;
	yrb = ytmp;
      }
      
      // Compute Pimp factor 
      double Pimp = 1.0;

	// Decide whether to use pT or QE for unordering suppression
      int uoType = 1;
      if (evolutionMode == 2) {
	// Note: if using different evolution types for different 
	// branchings, this is the one for the most recent branching
	uoType = antennaSets[0]->getAnt(antIndx)->getEvolutionType();
      }
      double Q2Hat = RPtr->Q2E(pClu[1],pClu[2],pClu[3],uoType);
      for (int jClu=1; jClu<int(pClu.size())-3; jClu++) {
	int jaHat=jClu+1;
	int jrHat=jClu+2;
	int jbHat=jClu+3;
	Q2Hat = min(Q2Hat,
		    RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],uoType));
      }	  
      double Q2New = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],uoType);
      Pimp = Q2Hat/(Q2Hat+Q2New);
      if (verbose >= 8) 
	printOut("Functions::getP2to4","Pimp = "+num2str(Pimp));
      
      // Compute ARIADNE factor (for g->qqbar splittings)
      double Pari = 1.0;
      if (antIndx == iqgSplit || antIndx == iggSplit) {
	double sN;
	if (!swapAB) {
	  // Only apply if neighbor is in final state
	  //	  if (ir+1 < int(pClu.size()) ) { 
	  sN = m2(pClu[ir],pClu[ir+1]);
	  // Only apply if neighbor is a quark
	  if (idClu[ir+1] != 21) 
	    Pari = 2.0*sN/(sP + sN);
	  //	  }
	} else { 
	  // Only apply if neighbor is in final state
	  //	  if (ir-2 >= nIn) {
	  sN = m2(pClu[ir-2],pClu[ir-1]);	
	  // Only apply if neighbor is a quark
	  if (idClu[ir-2] != 21)  
	    Pari = 2.0*sN/(sP + sN);
	  //	  }
	}
	
	// Verbose output
	if (verbose >= 8) 
	  printOut("Functions::getP2to4","Pari = "+num2str(Pari));
	
      }
      
      // Total prefactor
      double comFac = Pimp * Pari;
      
      // Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
      double Mp2 = 0.0 ; // mgInterfacePtr->ME2(nIn, idClu, pClu);
      
      // TMP!
      if (verbose >= 9) {
	cout << " P2to4 Mp2 = "<<Mp2<<endl;
	
	// Test
	double sC01  = m2(pClu[1],pClu[2]);
	double sC12  = m2(pClu[2],pClu[3]);
	double sC    = m2(pClu[1],pClu[2],pClu[3]);
	double A3chk = CF*antennaSets[0]->getAnt(iqqEmit)->aBarY(sC01/sC,sC12/sC)/sC;
	double A3chk2 = antennaSets[0]->getAnt(iqqEmit)->getChargeFactor()
	  *antennaSets[0]->getAnt(iqqEmit)->aBarY(sC01/sC,sC12/sC)/sC;
	cout << " P2to4 A30 = "<<scientific<<A3chk<<" s = "<<" A30chk = "
	     << A3chk2<<endl;
      }
      
      // Check if return was -1 ?
      if (abs(Mp2+1.0) < TINY) return false; 
      
      // Loop over different antenna sets (for uncertainty evaluation)
      for (int iVar=0; iVar<nVariations; iVar++) {
	
	// Compute physical antenna for this set for this (n-1)-parton state
	int iSet = 0;
	if (iVar >= 1) iSet = 1;
	if (iVar == 4) iSet = 2;
	if (iVar == 5) iSet = 3;
	double antPhys = antennaSets[iSet]->getAnt(antIndx)->aBarY(yar,yrb)/sP;
	double chargeFactor = antennaSets[iSet]->getAnt(antIndx)->getChargeFactor();
	if (iVar == 10 && id[ir] == 21) chargeFactor = CA;
	if (iVar == 11 && id[ir] == 21) chargeFactor = CF;
	antPhys *= chargeFactor;
	antPhys *= comFac;
	// Add to Approximated Mn2 for this set
	Mn2Appr[iVar] += Mp2 * antPhys;

     }
      
    }
    
    // Store ME reweight factors
    for (int iSet=0; iSet<int(antennaSets.size()); iSet++) {
      if (iSet >= 1 && ! uncertaintyBands) continue;
      //P2to4.push_back(Mn2/Mn2Appr[iSet]);
    }
    
    // Store Old ME
    //oldME2 = Mn2;
    
    
    // Done
    done = true;
    
    // For the time being, no non-MadGraph matching implemented
    return done;

  }

}
 
