// vincia04-root is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study n-jet rates and the so-called
// 4-jet angles LEP 1, using VINCIA's ROOT-based runtime display. 
// Uses VINCIA for showering and matching, FASTJET for jet clustering, 
// and PYTHIA 8 for all else
// Copyright (c) 2011 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
// FASTJET
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Auxiliary routine to compute angles between jets
double jetAngle(fastjet::PseudoJet jet1, fastjet::PseudoJet jet2) {
    Vec4 p1(jet1.px(),jet1.py(),jet1.pz(),jet1.e());
    Vec4 p2(jet2.px(),jet2.py(),jet2.pz(),jet2.e());
    return theta(p1,p2);
};

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia04.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event& event = pythia.event;

  // Some useful pointers
  Settings* settingsPtr = vincia.settingsPtr;        // to settings database
  //VinClu* vinCluPtr = &vincia.vinClu;                // to Vincia's 3->2 clu
  //Matching* matchingPtr = &vincia.matching;          // to Vincia's matching
  //MGInterface* mgInterfacePtr = &vincia.mgInterface; // to the MG interface

  // Extract settings to be used in the main program.  
  double eCM       = settingsPtr->parm("Beams:eCM");
  int    nEvent    = settingsPtr->mode("Main:numberOfEvents");
  int    nShow     = settingsPtr->mode("Main:timesToShow");

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  string antennaFile      = settingsPtr->word("Vincia:antennaFile");
  bool   uncertaintyBands = vinciaOn && settingsPtr->flag("Vincia:uncertaintyBands");
  int    verbose          = settingsPtr->mode("Vincia:verbose");
  int    orderingType     = settingsPtr->mode("Vincia:evolutionMode");
  int    evolutionType    = settingsPtr->mode("Vincia:evolutionType");
  int    showerNLC        = settingsPtr->mode("Vincia:NLC");

  // Extrac matching settings
  int    maxMatchTree     = settingsPtr->mode("VinciaMatching:order");
  bool   matching         = maxMatchTree >= 1;
  bool   scaleMatching    = settingsPtr->mode("Vincia:NLL") >= 1;
  int    matchingRegOrder = settingsPtr->mode("VinciaMatching:regOrder");
  // Example of extracting a mass value:
  // double bMass            = vincia.particleDataPtr->m0(5);

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot object 
  VinciaRoot vroot; 

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose(verbose);
  // Switch stats on/off (in ROOT conventions, but with 0 = off)
  vroot.setOptStat(0);

  // Set MC names
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  string MCnames;
  if (vinciaOn) {
    MCnames += "Vincia "+vcVer+" +";
    if (matching) MCnames += " MadGraph 4.426 + ";
  }
  MCnames += " Pythia "+pyVer;
  vroot.setMCnames(MCnames);

  // Set label for plot legend
  string mcLabel = "Vincia";  
  if (!vinciaOn) mcLabel = "Pythia";
  vroot.setLegend(mcLabel);

  // Set condensed MC name for saving to dat files
  string mcFile = "vc";  
  if (!vinciaOn) mcFile = "py";
  if (!settingsPtr->flag("HadronLevel:all")) mcFile+="-nohad";

  // Optionally set name for saved dat file(s) to read in and overplot
  // Compare Global Vincia to Pythia
  string mcOld = "";    
  if (vinciaOn) {
    mcOld = "py";
  }
  if (!settingsPtr->flag("HadronLevel:all")) mcOld+="-nohad";

  // Set names for uncertainty bands, if using them (for plot legends)
  /*
  if (uncertaintyBands) {
    vroot.addUncertainty("Def",0,1);
    vroot.addUncertainty("#muR",2,3);
    vroot.addUncertainty("Finite",4,5);
    vroot.addUncertainty("QMatch",6,7);
    vroot.addUncertainty("Ord",8,9);
    vroot.addUncertainty("1/N_{C}^{2}",10,11);
  }
  */

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythia.init();

  // List changed settings
  pythia.settings.listChanged();

  // Check that Z0 decay channels set correctly.
  pythia.particleData.listChanged();
 
  //***************************************************************************
  // FASTJET
  //***************************************************************************

  // Fastjet analysis - select algorithm and parameters
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::ee_kt_algorithm,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;

  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  if (not vinciaOn) {
    outfile<<"pythia04";
  } else {
    outfile<<"vincia04";
  }
  if (settingsPtr->flag("HadronLevel:all")) outfile<<"-had";
  outfile<<".root";

  //****************
  // Book Histograms

  // Vincia uncertainty bands: number of variations expected 
  int nVariations = max(1,vincia.nVariations());

  // Jet Rates
  int nBins  = 6;
  double xLo = -1.2;
  double xHi =  0.0;
  vector<TH1D*> Vr1j;
  Vr1j.push_back(new TH1D("r1j","Exclusive 1-Jet Rate",nBins, xLo, xHi));
  // Add to display with
  // logy = true, normalize=false, showStatErr=true
  Vr1j[0]->SetMaximum(1.01); 
  Vr1j[0]->SetMinimum(0.00004); 
  Vr1j[0]->GetXaxis()->SetTitle("Ln(y1)");
  Vr1j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr1j.push_back(new TH1D(string("r1j_"+num2str(iVar,1)).c_str()," ",
                            nBins,xLo,xHi));  
  vroot.createDisplay("r1j",Vr1j,320,400);
  vroot.setDataFile("r1j","data/Aleph-91-r1j.vec",10,30,40,51,52,61,62,71,72);
  vroot.addHisFile("r1j",mcOld+"-r1j.dat");
  vroot.setNormalize("r1j",false);
  vroot.setLogY("r1j");

  // Jet Rates
  nBins = 51;
  xLo   = -10.2;
  xHi   =   0.0;
  vector<TH1D*> Vr2j;
  Vr2j.push_back(new TH1D("r2j","Exclusive 2-Jet Rate",nBins,xLo,xHi));
  Vr2j[0]->SetMaximum(200.01); 
  Vr2j[0]->SetMinimum(0.00004); 
  Vr2j[0]->GetXaxis()->SetTitle("Ln(y2)");
  Vr2j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr2j.push_back(new TH1D(string("r2j_"+num2str(iVar,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  vroot.createDisplay("r2j",Vr2j,320,400);
  vroot.setDataFile("r2j","data/Aleph-91-r2j.vec",10,30,40,51,52,61,62,71,72);
  vroot.addHisFile("r2j",mcOld+"-r2j.dat");
  vroot.setNormalize("r2j",false); 
  vroot.setLogY("r2j");

  // Jet Rates
  nBins =    46;
  xLo   = -10.2;
  xHi   =  -1.0;
  vector<TH1D*> Vr3j;
  Vr3j.push_back(new TH1D("r3j","Exclusive 3-Jet Rate",nBins,xLo,xHi));
  Vr3j[0]->SetMaximum(200.01); 
  Vr3j[0]->SetMinimum(0.00002); 
  Vr3j[0]->GetXaxis()->SetTitle("Ln(y3)");
  Vr3j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr3j.push_back(new TH1D(string("r3j_"+num2str(iVar,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  vroot.createDisplay("r3j",Vr3j,320,400);
  vroot.setDataFile("r3j","data/Aleph-91-r3j.vec",10,30,40,51,52,61,62,71,72);
  vroot.addHisFile("r3j",mcOld+"-r3j.dat");
  vroot.setNormalize("r3j",false); 
  vroot.setLogY("r3j");

  // R3bl 
  double nB(0), nC(0), nL(0);
  int nBinsR3bl = 44;
  xLo = 0.00;
  xHi = 0.055;
  double binWidR3bl = (xHi-xLo)/nBinsR3bl; 
  vector<TH1D*> VR3bl;
  VR3bl.push_back(new TH1D("R3bl","R3bl",nBinsR3bl, xLo, xHi));
  VR3bl[0]->SetMaximum(1.11); 
  VR3bl[0]->SetMinimum(0.89); 
  VR3bl[0]->GetXaxis()->SetTitle("y_{23}");
  VR3bl[0]->GetYaxis()->SetTitle("R3bl");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VR3bl.push_back(new TH1D(string("R3bl_"+num2str(iVar,1)).c_str(),
                                            " ",nBinsR3bl,xLo,xHi));  
  vroot.createDisplay("R3bl",VR3bl,320,400);
  vroot.setDataFile("R3bl","data/Delphi-91-R3bl-Durham-had.vec",
                    20,40,50,60);
  vroot.addHisFile("R3bl",mcOld+"-R3bl.dat");
  vroot.setLogY("R3bl",false);
  vroot.setNormalize("R3bl",false);

  // Jet Rates
  nBins =    41;
  xLo   = -10.2;
  xHi   =  -2.0;
  vector<TH1D*> Vr4j;
  Vr4j.push_back(new TH1D("r4j","Exclusive 4-Jet Rate",nBins,xLo,xHi));
  Vr4j[0]->SetMaximum(120.01); 
  Vr4j[0]->SetMinimum(0.000005); 
  Vr4j[0]->GetXaxis()->SetTitle("Ln(y4)");
  Vr4j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr4j.push_back(new TH1D(string("r4j_"+num2str(iVar,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  vroot.createDisplay("r4j",Vr4j,320,400);
  vroot.setDataFile("r4j","data/Aleph-91-r4j.vec",10,30,40,51,52,61,62,71,72);
  vroot.addHisFile("r4j",mcOld+"-r4j.dat");
  vroot.setNormalize("r4j",false); 
  vroot.setLogY("r4j");

  // Jet Rates
  nBins =    36;
  xLo   = -10.2;
  xHi   =  -3.0;
  vector<TH1D*> Vr5j;
  Vr5j.push_back(new TH1D("r5j","Exclusive 5-Jet Rate",nBins,xLo,xHi));
  Vr5j[0]->SetMaximum(20.01); 
  Vr5j[0]->SetMinimum(0.000005); 
  Vr5j[0]->GetXaxis()->SetTitle("Ln(y5)");
  Vr5j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr5j.push_back(new TH1D(string("r5j_"+num2str(iVar,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  vroot.createDisplay("r5j",Vr5j,320,400);
  vroot.setDataFile("r5j","data/Aleph-91-r5j.vec",10,30,40,51,52,61,62,71,72);
  vroot.addHisFile("r5j",mcOld+"-r5j.dat");
  vroot.setNormalize("r5j",false); 
  vroot.setLogY("r5j");

  // Jet Rates
  nBins =    31;
  xLo   = -10.2;
  xHi   =  -4.0;
  vector<TH1D*> Vr6j;
  Vr6j.push_back(new TH1D("r6j","Inclusive 6-Jet Rate",nBins,xLo,xHi));
  Vr6j[0]->SetMaximum(20.01); 
  Vr6j[0]->SetMinimum(0.000005); 
  Vr6j[0]->GetXaxis()->SetTitle("Ln(y6)");
  Vr6j[0]->GetYaxis()->SetTitle("Rate");
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vr6j.push_back(new TH1D(string("r6j_"+num2str(iVar,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  vroot.createDisplay("r6j",Vr6j,320,400);
  vroot.setDataFile("r6j","data/Aleph-91-r6j.vec",10,30,40,51,52,61,62,71,72);
  vroot.addHisFile("r6j",mcOld+"-r6j.dat");
  vroot.setNormalize("r6j",false); 
  vroot.setLogY("r6j");

  // Jet Resolutions
  nBins = 13;
  xLo   = 0.0;
  xHi   = 1.3;
  double wY12 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY12;
  VLnY12.push_back(new TH1D("LnY12","Ln(1/y12)",nBins, xLo, xHi));
  VLnY12[0]->SetMaximum(20.1); 
  VLnY12[0]->SetMinimum(0.000005); 
  VLnY12[0]->GetXaxis()->SetTitle("Ln(1/y12)");
  VLnY12[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
     VLnY12.push_back(new TH1D(string("LnY12_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  vroot.createDisplay("LnY12",VLnY12,320,400,400,0);
  vroot.setDataFile("LnY12","data/Aleph-91-LnY12.vec",10,30,40,50,60,70);
  vroot.addHisFile("LnY12",mcOld+"-LnY12.dat");
  vroot.setLogY("LnY12");

  nBins = 51;
  xLo = 0.8;
  xHi = 11.0;
  double wY23 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY23;
  VLnY23.push_back(new TH1D("LnY23","Ln(1/y23)",nBins, xLo, xHi));
  VLnY23[0]->SetMaximum(20.01); 
  VLnY23[0]->SetMinimum(0.00005); 
  VLnY23[0]->GetXaxis()->SetTitle("Ln(1/y23)");
  VLnY23[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VLnY23.push_back(new TH1D(string("LnY23_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  vroot.createDisplay("LnY23",VLnY23,320,400,416,16);
  vroot.setDataFile("LnY23","data/Aleph-91-LnY23.vec",10,30,40,50,60,70);
  vroot.addHisFile("LnY23",mcOld+"-LnY23.dat");
  vroot.setLogY("LnY23");

  nBins = 48;
  xLo = 1.8;
  xHi = 11.4;
  double wY34 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY34;
  VLnY34.push_back(new TH1D("LnY34","Ln(1/y34)",nBins, xLo, xHi));
  VLnY34[0]->SetMaximum(20.01); 
  VLnY34[0]->SetMinimum(0.00005); 
  VLnY34[0]->GetXaxis()->SetTitle("Ln(1/y34)");
  VLnY34[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VLnY34.push_back(new TH1D(string("LnY34_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi)); 
  vroot.createDisplay("LnY34",VLnY34,320,400,432,32);
  vroot.setDataFile("LnY34","data/Aleph-91-LnY34.vec",10,30,40,50,60,70);
  vroot.addHisFile("LnY34",mcOld+"-LnY34.dat");
  vroot.setLogY("LnY34");

  nBins = 45;
  xLo = 3.0;
  xHi = 12.0;
  double wY45 = 200.0/7.0;
  vector<TH1D*> VLnY45;
  VLnY45.push_back(new TH1D("LnY45","Ln(1/y45)",nBins, xLo, xHi));
  VLnY45[0]->SetMaximum(20.01); 
  VLnY45[0]->SetMinimum(0.00005); 
  VLnY45[0]->GetXaxis()->SetTitle("Ln(1/y45)");
  VLnY45[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VLnY45.push_back(new TH1D(string("LnY45_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  vroot.createDisplay("LnY45",VLnY45,320,400,448,48);
  vroot.setDataFile("LnY45","data/Aleph-91-LnY45.vec",10,30,40,50,60,70);
  vroot.addHisFile("LnY45",mcOld+"-LnY45.dat");
  vroot.setLogY("LnY45");

  nBins = 46;
  xLo = 3.8;
  xHi = 13.0;
  double wY56 = 200.0/6.;
  vector<TH1D*> VLnY56;
  VLnY56.push_back(new TH1D("LnY56","Ln(1/y56)",nBins, xLo, xHi));
  VLnY56[0]->SetMaximum(20.01); 
  VLnY56[0]->SetMinimum(0.00005); 
  VLnY56[0]->GetXaxis()->SetTitle("Ln(1/y56)");
  VLnY56[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VLnY56.push_back(new TH1D(string("LnY56_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  vroot.createDisplay("LnY56",VLnY56,320,400,464,64);
  vroot.setDataFile("LnY56","data/Aleph-91-LnY56.vec",10,30,40,50,60,70);
  vroot.addHisFile("LnY56",mcOld+"-LnY56.dat");
  vroot.setLogY("LnY56");

  // BZ Angle
  nBins = 20;
  xLo   = 0.0;
  xHi   = 1.0;
  double wBZ = 20.0/1.0;
  vector<TH1D*> VChiBZ;
  VChiBZ.push_back(new TH1D("CosChiBZ","Bengtsson-Zerwas Angle",nBins, xLo, xHi));
  VChiBZ[0]->SetMaximum(4.0); 
  VChiBZ[0]->SetMinimum(-0.2); 
  VChiBZ[0]->GetXaxis()->SetTitle("cos(#Chi_{BZ})");
  VChiBZ[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
                  VChiBZ.push_back(new TH1D(string("ChiBZ_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  vroot.createDisplay("ChiBZ",VChiBZ,320,400,16,216);
  vroot.setDataFile("ChiBZ","data/Delphi-91-chiBZ.vec",
                    10,30,40,50,60);
  vroot.addHisFile("ChiBZ",mcOld+"-ChiBZ.dat");

  // KSW Angle
  nBins = 20;
  xLo   = -1.0;
  xHi   =  1.0;
  double wKSW = 20.0/2.0;
  vector<TH1D*> VPhiKSW;
  VPhiKSW.push_back(new TH1D("CosPhiKSW","Korner-Schierholz-Willrodtorner Angle",
                             nBins, xLo, xHi));
  VPhiKSW[0]->SetMaximum(1.5); 
  VPhiKSW[0]->SetMinimum(-0.075); 
  VPhiKSW[0]->GetXaxis()->SetTitle("cos(#Phi_{KSW})");
  VPhiKSW[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VPhiKSW.push_back(new TH1D(string("PhiKSW_"+num2str(iVar,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  vroot.createDisplay("PhiKSW",VPhiKSW,320,400,32,232);
  vroot.setDataFile("PhiKSW","data/Delphi-91-phiKSW.vec",
                    10,30,40,50,60);
  vroot.addHisFile("PhiKSW",mcOld+"-PhiKSW.dat");

  // NR Angle
  nBins = 20;
  xLo   =  0.0;
  xHi   =  1.0;
  double wNR = 20.0/1.0;
  vector<TH1D*> VThetaNR;
  VThetaNR.push_back(new TH1D("CosThetaNR","Nachtman-Reiter Angle", nBins, xLo, xHi));
  VThetaNR[0]->SetMaximum(2.5); 
  VThetaNR[0]->SetMinimum(-0.125); 
  VThetaNR[0]->GetXaxis()->SetTitle("|cos(#Theta_{NR})|");
  VThetaNR[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VThetaNR.push_back(new TH1D(string("ThetaNR_"+num2str(iVar,1)).c_str(),
                                  " ",nBins,xLo,xHi));  
  vroot.createDisplay("ThetaNR",VThetaNR,320,400,48,248);
  vroot.setDataFile("ThetaNR","data/Delphi-91-thetaNR.vec",
                    10,30,40,50,60);
  vroot.addHisFile("ThetaNR",mcOld+"-ThetaNR.dat");

  // 34 Angle
  nBins = 20;
  xLo   = -1.0;
  xHi   = 1.0;
  double wA34 = 20.0/2.0;  
  vector<TH1D*> VAlpha34;
  VAlpha34.push_back(new TH1D("CosAlpha34","#alpha_{34}", nBins, xLo, xHi));
  VAlpha34[0]->SetMaximum(1.2); 
  VAlpha34[0]->SetMinimum(-0.06); 
  VAlpha34[0]->GetXaxis()->SetTitle("cos(#alpha_{34})");
  VAlpha34[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iVar=1; iVar<nVariations; iVar++) 
      VAlpha34.push_back(new TH1D(string("Alpha34_"+num2str(iVar,1)).c_str(),
                                  " ",nBins,xLo,xHi));  
  vroot.createDisplay("Alpha34",VAlpha34,320,400,64,264);
  vroot.setDataFile("Alpha34","data/Delphi-91-a34.vec",
                    10,30,40,50,60);
  vroot.addHisFile("Alpha34",mcOld+"-Alpha34.dat");

  // nCharged
  TH1D *HnCharged=new TH1D("nCh","nCharged", 60, -0.5, 59.5);
  HnCharged->SetMinimum(0.00001);
  HnCharged->SetMaximum(0.5);

  // Ln(1/x)
  TH1D *Hx=new TH1D("LnXInv","Ln(1/x)", 200, 0.0, 5.4);
  double wx = 200.0/5.4;
  //vroot.addTH1DPtr(Hx,"Fragmentation",true,true,true,"data/Aleph-91-LogInvx.vec");
  Hx->SetMinimum(0.0002);
  Hx->SetMaximum(8.);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtr->mode("Random:seed");
  vincia.rndmPtr->init(initseed);

  // Counter for negative-weight events
  double evwt=1.0;
  double sumw[nVariations];
  for (int iVar=0; iVar<nVariations; iVar++) sumw[iVar]=0.0;

  // Counters for jet rates
  nBins = 100;
  double sumN1j[nBins][nVariations];  
  double sumN2j[nBins][nVariations];
  double sumN3j[nBins][nVariations];
  double sumN4j[nBins][nVariations];
  double sumN5j[nBins][nVariations];
  double sumN6j[nBins][nVariations];
  double sumY23l[nBinsR3bl][nVariations];
  double sumY23b[nBinsR3bl][nVariations];
  for (int iBin=0; iBin<nBins; iBin++) 
    for (int iVar=0; iVar<nVariations; iVar++) {
      sumN1j[iBin][iVar] = 0.0;
      sumN2j[iBin][iVar] = 0.0;
      sumN3j[iBin][iVar] = 0.0;
      sumN4j[iBin][iVar] = 0.0;
      sumN5j[iBin][iVar] = 0.0;
      sumN6j[iBin][iVar] = 0.0;
      if (iBin < nBinsR3bl) {
        sumY23l[iBin][iVar] = 0.0;
        sumY23b[iBin][iVar] = 0.0;
      }
    }
  

  // FastJet needs bool to tell whether this event is the first one
  bool firstEvent = true; 

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {


    if (iEvent%(max(1,nEvent/max(nShow,1))) == 0)  
      cout << " Now begin event " << iEvent << endl;

    if (!pythia.next()) continue;

    // Event weight
    if (vinciaOn) evwt    = vincia.weight();
    for (int iVar=0; iVar<nVariations; iVar++) 
      sumw[iVar] += vincia.weight(iVar);

    // List the first few events
    if (iEvent < 3) event.list();

    // Count FS charged hadrons, partons, and quarks
    bool isZ2B   = (event[6].idAbs() == 5);
    bool isZ2C   = (event[6].idAbs() == 4);
    int  nCh     = 0;

    // Add final-state particles to FastJet input vector    
    fjInputs.resize(0);
    for (int i=1;i<event.size();i++) {

      // Final state only
      if (!pythia.event[i].isFinal())        continue;
      
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) {
        nCh++;
        double LnXInv=log(double(1.0/(2.0*event[i].pAbs()/eCM)));
        Hx->Fill(LnXInv,evwt*wx);
      }

      // Now decide what to pass to FastJet
      // NOTE: neutrinos are *included* in this list, since ALEPH
      // appears to have corrected their data to include them.
      // Store as input to Fastjet
      fjInputs.push_back( fastjet::PseudoJet (pythia.event[i].px(),
                                              pythia.event[i].py(), 
                                              pythia.event[i].pz(),
                                              pythia.event[i].e() ) );

    }

    // Counter for B events
    if (isZ2B) nB++;
    else if (isZ2C) nC++;
    else nL++;

    // FastJet
    if (fjInputs.size() == 0) {
      cout << "Error: event with no final state particles" << endl;
      continue;
    }

    // Run Fastjet algorithm    
    fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);

    // 4-jet angles Use yCut from LEP 4-jet angle analysis
    vector<fastjet::PseudoJet> fastjets =  clustSeq.exclusive_jets_ycut(0.008);
    vector<fastjet::PseudoJet> sorted = fastjet::sorted_by_E(fastjets);

    // Extract merging scales
    double y56   = clustSeq.exclusive_ymerge(5);
    double lny56 = log(y56);
    double y45   = clustSeq.exclusive_ymerge(4);
    double lny45 = log(y45);
    double y34   = clustSeq.exclusive_ymerge(3);
    double lny34 = log(y34);
    double y23   = clustSeq.exclusive_ymerge(2);
    double lny23 = log(y23);
    double lny12 = log(clustSeq.exclusive_ymerge(1));

    // Fill Jet resolution histograms with results from FastJet
    for (int iVar=0; iVar < nVariations; iVar++) {
      VLnY12[iVar]->Fill(-lny12,wY12*vincia.weight(iVar));
      VLnY23[iVar]->Fill(-lny23,wY23*vincia.weight(iVar));
      VLnY34[iVar]->Fill(-lny34,wY34*vincia.weight(iVar));
      VLnY45[iVar]->Fill(-lny45,wY45*vincia.weight(iVar));
      VLnY56[iVar]->Fill(-lny56,wY56*vincia.weight(iVar));
      // R3bl (used in paper on massive antennae)
      int iy = -1;
      for (double y=0.0+binWidR3bl/2.; y<=0.055; y += binWidR3bl) {
        ++iy;
        // Check if event looks like 3-jet event at this scale
        if (y23 >= y && iy < nBinsR3bl) {
          if (isZ2B) sumY23b[iy][iVar] += vincia.weight(iVar);
          else if (!isZ2C) sumY23l[iy][iVar] += vincia.weight(iVar);
        }
      }
      VR3bl[iVar]->Reset();
      iy = -1;
      for (double y=0.0+binWidR3bl/2.; y<=0.055; y += binWidR3bl) {
        ++iy;
        double valb  = sumY23b[iy][iVar] / max(1.0,nB); 
        double errb  = sqrt(sumY23b[iy][iVar]) / max(1.0,nB);
        double vall  = sumY23l[iy][iVar] / max(1.0,nL);
        double errl  = sqrt(sumY23l[iy][iVar]) / max(1.0,nL);
        double R3bl = 1.0;
        double errR = 0.0;
        if (vall > 0 && valb > 0) {
          R3bl = valb/vall;
          errR = sqrt(pow2(errb/valb) + pow2(errl/vall)) * R3bl;
        }
         int iBin    = VR3bl[iVar]->FindBin(y);
        // Don't fill very first bin where there is no data (restricts y axis)
        if (iBin >= 3) {
          VR3bl[iVar]->SetBinContent(iBin,R3bl);        
          VR3bl[iVar]->SetBinError(iBin,errR);        
        }
      }

    }


    // Histogram jet rates
    // Loop over resolutions
    for (int iVar=0; iVar < nVariations; iVar++) {
      Vr1j[iVar]->Reset();
      Vr2j[iVar]->Reset();
      Vr3j[iVar]->Reset();
      Vr4j[iVar]->Reset();
      Vr5j[iVar]->Reset();
      Vr6j[iVar]->Reset();
      int iy=-1;
      for (double lny=-10.1; lny<0.0; lny += 0.2) {
        iy++;
        if (lny12 < lny) sumN1j[iy][iVar] += vincia.weight(iVar);
        else if (lny23 < lny) sumN2j[iy][iVar] += vincia.weight(iVar);
        else if (lny34 < lny) sumN3j[iy][iVar] += vincia.weight(iVar);
        else if (lny45 < lny) sumN4j[iy][iVar] += vincia.weight(iVar);
        else if (lny56 < lny) sumN5j[iy][iVar] += vincia.weight(iVar);
        else  sumN6j[iy][iVar] += vincia.weight(iVar);
        double err, val;
        // 1j
        val = sumN1j[iy][iVar]/sumw[iVar];
        err = sqrt(sumN1j[iy][iVar])/sumw[iVar];
        int iBin = Vr1j[iVar]->FindBin(lny);
        Vr1j[iVar]->SetBinContent(iBin,val);
        Vr1j[iVar]->SetBinError(iBin,err);

        // 2j
        Vr2j[iVar]->Fill(lny,sumN2j[iy][iVar]/sumw[iVar]);
        err = sqrt(sumN2j[iy][iVar])/sumw[iVar];
        iBin = Vr2j[iVar]->FindBin(lny);
        Vr2j[iVar]->SetBinError(iBin,err);
        // 3j
        Vr3j[iVar]->Fill(lny,sumN3j[iy][iVar]/sumw[iVar]);
        err = sqrt(sumN3j[iy][iVar])/sumw[iVar];
        iBin = Vr3j[iVar]->FindBin(lny);
        Vr3j[iVar]->SetBinError(iBin,err);
        // 4j
        Vr4j[iVar]->Fill(lny,sumN4j[iy][iVar]/sumw[iVar]);
        err = sqrt(sumN4j[iy][iVar])/sumw[iVar];
        iBin = Vr4j[iVar]->FindBin(lny);
        Vr4j[iVar]->SetBinError(iBin,err);
        // 5j
        Vr5j[iVar]->Fill(lny,sumN5j[iy][iVar]/sumw[iVar]);
        err = sqrt(sumN5j[iy][iVar])/sumw[iVar];
        iBin = Vr5j[iVar]->FindBin(lny);
        Vr5j[iVar]->SetBinError(iBin,err);
        // 6j
        Vr6j[iVar]->Fill(lny,sumN6j[iy][iVar]/sumw[iVar]);
        err = sqrt(sumN6j[iy][iVar])/sumw[iVar];
        iBin = Vr6j[iVar]->FindBin(lny);
        Vr6j[iVar]->SetBinError(iBin,err);
      }
    }

    // Fill charged-particle histogram
    HnCharged->Fill(1.0*nCh,evwt);


    // Periodically update displays
    if (iEvent%500 == 499|| iEvent == nEvent-1) 
      vroot.updateDisplay();       
    if (iEvent == 1000 || iEvent%10000==0 || iEvent == nEvent-1) {
      string mcStr="vincia";
      if (not vinciaOn) mcStr="pythia";
      else {
        if (evolutionType != 0) mcStr += "-e"+num2str(evolutionType,1);
        if (antennaFile.find("Max") != string::npos) mcStr +="-max";
        else if (antennaFile.find("Min") != string::npos) mcStr +="-min";
        else if (antennaFile.find("Ari") != string::npos) mcStr +="-ari";
        //else mcStr += "-def";
        if (showerNLC <= 0) mcStr += "LC";
        if (not matching) {
          if (orderingType == 3) mcStr+="-m0";
          else if (orderingType == 1) mcStr+="-ord";
        }
        else {
          mcStr += "-m"+num2str(maxMatchTree,1);
          if (matchingRegOrder != 3) mcStr += "r"+num2str(maxMatchTree,1);
        }
        if (not scaleMatching) mcStr += "-mu0";
        if (uncertaintyBands) mcStr+="-U";
      }
      if (!settingsPtr->flag("HadronLevel:all")) mcStr+="-nohad";
      
      // Save plots
      vroot.saveDisplays(mcStr,"eps");

    }

    // Now only keep exclusive 4-jet events
    if (sorted.size() != 4) continue;

    // For the first event, print the FastJet details
    if (firstEvent) {
      cout << "Ran " << jetDef->description() << endl;
      cout << "Strategy adopted by FastJet was "
           << clustSeq.strategy_string() << endl << endl;      
      firstEvent = false;
    }

    // Compute 4-jet angles
    Vec4 p1(sorted[0].px(),sorted[0].py(),sorted[0].pz(),sorted[0].e());
    Vec4 p2(sorted[1].px(),sorted[1].py(),sorted[1].pz(),sorted[1].e());
    Vec4 p3(sorted[2].px(),sorted[2].py(),sorted[2].pz(),sorted[2].e());
    Vec4 p4(sorted[3].px(),sorted[3].py(),sorted[3].pz(),sorted[3].e());
    // Bengtsson-Zerwas
    Vec4 v1 = cross3(p1,p2);
    Vec4 v2 = cross3(p3,p4);
    double cosChiBZ = abs(costheta(v1,v2));
    // KSW
    v1 = cross3(p1,p4);
    v2 = cross3(p2,p3);
    double alpha1 = theta(v1,v2);
    v1 = cross3(p1,p3);
    v2 = cross3(p2,p4);
    double alpha2 = theta(v1,v2);
    double cosPhiKSW = cos((alpha1+alpha2)/2.);
    // Nachtman-Reiter
    v1 = p1 - p2;
    v2 = p3 - p4;
    double cosThetaNR = abs(costheta(v1,v2));
    // Alpha34
    double cosA34 = costheta(p3,p4);

    // Fill 4-jet angle histograms
    for (int iVar=0; iVar < nVariations; iVar++) {
      VChiBZ[iVar]->Fill(cosChiBZ,vincia.weight(iVar)*wBZ);
      VThetaNR[iVar]->Fill(cosThetaNR,vincia.weight(iVar)*wNR);
      VAlpha34[iVar]->Fill(cosA34,vincia.weight(iVar)*wA34);
      VPhiKSW[iVar]->Fill(cosPhiKSW,vincia.weight(iVar)*wKSW);
    }
  }

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.statistics();
  if (vinciaOn) vincia.printInfo();

  // Example of saving histograms to ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  TFile f(outfile.str().c_str(),"recreate");
  for (int iVar=0; iVar<nVariations; ++iVar) {
    VLnY12[iVar]->Write();
    VLnY23[iVar]->Write();
    VLnY34[iVar]->Write();
    VLnY45[iVar]->Write();
    VLnY56[iVar]->Write();
    VChiBZ[iVar]->Write();
    VThetaNR[iVar]->Write();
    VAlpha34[iVar]->Write();
    VPhiKSW[iVar]->Write();
  }
  f.Close();

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
