// vincia08-root is a part of the VINCIA plugin to PYTHIA 8
// This program runs three instances of PYTHIA 8 in parallel, one of which
// is standard PYTHIA 8, the others use two different VINCIA settings. 
// All 3 results are plotted simultaneously using the VinciaRoot interface
// Some observables corresponding to compressed hierarchies are computed.
// Copyright (c) 2011 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
// FASTJET
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Auxiliary routine to compute angles between jets
double jetAngle(fastjet::PseudoJet jet1, fastjet::PseudoJet jet2) {
    Vec4 p1(jet1.px(),jet1.py(),jet1.pz(),jet1.e());
    Vec4 p2(jet2.px(),jet2.py(),jet2.pz(),jet2.e());
    return theta(p1,p2);
};

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythiaA; 
  Pythia pythiaB; 
  Pythia pythiaC; 
  Pythia pythiaD; 
  VinciaPlugin vinciaA(&pythiaA,"vincia08a.cmnd");
  VinciaPlugin vinciaB(&pythiaB,"vincia08b.cmnd");
  VinciaPlugin vinciaC(&pythiaC,"vincia08c.cmnd");
  VinciaPlugin vinciaD(&pythiaD,"vincia08d.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Pointers to A and B settings
  Settings* settingsPtrA = vinciaA.settingsPtr;     
  Settings* settingsPtrB = vinciaB.settingsPtr;     
  Settings* settingsPtrC = vinciaC.settingsPtr;     
  Settings* settingsPtrD = vinciaD.settingsPtr;     
  bool   vinciaOnA       = settingsPtrA->flag("Vincia");
  bool   vinciaOnB       = settingsPtrB->flag("Vincia");
  bool   vinciaOnC       = settingsPtrC->flag("Vincia");
  bool   vinciaOnD       = settingsPtrD->flag("Vincia");

  // Extract settings to be used in the main program.  
  double eCMa      = settingsPtrA->parm("Beams:eCM");
  double eCMb      = settingsPtrB->parm("Beams:eCM");
  double eCMc      = settingsPtrC->parm("Beams:eCM");
  double eCMd      = settingsPtrD->parm("Beams:eCM");
  int    nEvent    = settingsPtrA->mode("Main:numberOfEvents");
  int    nShow     = settingsPtrA->mode("Main:timesToShow");
  int    verbose   = settingsPtrB->mode("Vincia:verbose");

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot object 
  VinciaRoot vroot; 
  vroot.setLogY();

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose(verbose);
  // Switch stats on/off (in ROOT conventions, but with 0 = off)
  vroot.setOptStat(0);
  // Tell it the name of the MC we are running (for plot legends)
  string pyVer    = num2str(settingsPtrB->parm("Pythia:versionNumber"),5);  
  string vcVer    = num2str(settingsPtrB->parm("Vincia:versionNumber"),5);  

  // Tell it the name of the MC we are running (for plot legends)
  vector<string> legend;
  legend.push_back("Pythia");
  legend.push_back("Vincia (smooth)");
  legend.push_back("Vincia (strong)");
  legend.push_back("Vincia (mD,str)");
  vroot.setLegend(legend);

  string MCnames = "Vincia "+vcVer;
  MCnames += " + MadGraph 4.426";
  MCnames += " + Pythia "+pyVer;
  vroot.setMCnames(MCnames);

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythiaA.init( 11, -11, eCMa);
  pythiaB.init( 11, -11, eCMb);
  pythiaC.init( 11, -11, eCMc);
  pythiaD.init( 11, -11, eCMc);

  // List changed settings
  pythiaB.settings.listChanged();

  // Check that Z0 decay channels set correctly.
  pythiaB.particleData.listChanged();
 
  //***************************************************************************
  // FASTJET
  //***************************************************************************

  // Fastjet analysis - select algorithm and parameters
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::ee_kt_algorithm,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;

  //***************************************************************************
  // ROOT
  //***************************************************************************

  //****************
  // Book Histograms
  double xLo, xHi;
  int nBins;

  // Jet resolution ratios:
  // Want to probe perturbative physics above scale for B decay, so 
  // restrict region of interest to 
  //   all yij > 0.005 ~=> ln(yij) > -5.3   ~=>   kT > 6.5 GeV
  // Also require y(ij-n) > 0.5 y(23)   => compressed hierarchies
  // Implies y(23) > 0.01   =>   ln(y23) > -4.6

  // Same Side LnY24
  nBins = 10;
  xLo = -0.7;
  xHi =  0.3;
  double wLnY24SS = nBins/(xHi-xLo);
  vector<TH1D*> VLnY24SS;
  VLnY24SS.push_back(new TH1D("LnY24SS","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS.push_back(new TH1D("LnY24SSb","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS.push_back(new TH1D("LnY24SSc","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS.push_back(new TH1D("LnY24SSd","SameSide Ln(y34/y23)",nBins, xLo, xHi));
  VLnY24SS[0]->SetMaximum(1.0); 
  VLnY24SS[0]->SetMinimum(0.005); 
  VLnY24SS[0]->GetXaxis()->SetTitle("#xi_{24}");
  VLnY24SS[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("LnY24SS",350,350);
  for (unsigned int iHis=0; iHis < VLnY24SS.size(); ++iHis) {
    vroot.addTH1DPtr("LnY24SS",VLnY24SS[iHis]);  
  }

  // Same Side LnY25
  nBins =  5;
  xLo = -1.0;
  xHi =  0.25;
  double wLnY25SS = nBins/(xHi-xLo);
  vector<TH1D*> VLnY25SS;
  VLnY25SS.push_back(new TH1D("LnY25SS","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS.push_back(new TH1D("LnY25SSb","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS.push_back(new TH1D("LnY25SSc","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS.push_back(new TH1D("LnY25SSd","SameSide Ln(y45/y23)",nBins, xLo, xHi));
  VLnY25SS[0]->SetMaximum(0.5); 
  VLnY25SS[0]->SetMinimum(0.001); 
  VLnY25SS[0]->GetXaxis()->SetTitle("#xi_{25}");
  VLnY25SS[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("LnY25SS",350,350);
  for (unsigned int iHis=0; iHis < VLnY25SS.size(); ++iHis) {
    vroot.addTH1DPtr("LnY25SS",VLnY25SS[iHis]);  
  }

  // Jet mass ratio
  nBins =   20;
  xLo   =  0.0;
  xHi   =  1.0;
  double wDM = nBins/(xHi-xLo);
  vector<TH1D*> VDM;
  VDM.push_back(new TH1D("DM","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM.push_back(new TH1D("DMb","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM.push_back(new TH1D("DMc","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM.push_back(new TH1D("DMd","M^{2}_{L}/M^{2}_{H} with cuts",nBins, xLo, xHi));
  VDM[0]->SetMaximum(2.5); 
  VDM[0]->SetMinimum(0.0); 
  VDM[0]->GetXaxis()->SetTitle("M^{2}_{L}/M^{2}_{H}");
  VDM[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("DM",350,350);
  for (unsigned int iHis=0; iHis < VDM.size(); ++iHis) {
    vroot.addTH1DPtr("DM",VDM[iHis]);  
  }

  // SameSide fractions 
  nBins =   3;
  xLo   =  1.5;
  xHi   =  4.5;
  vector<TH1D*> VSS;
  VSS.push_back(new TH1D("SS","SameSide Fractions",nBins, xLo, xHi));
  VSS.push_back(new TH1D("SSb","SameSide Fractions",nBins, xLo, xHi));
  VSS.push_back(new TH1D("SSc","SameSide Fractions",nBins, xLo, xHi));
  VSS.push_back(new TH1D("SSd","SameSide Fractions",nBins, xLo, xHi));
  VSS[0]->SetMaximum(0.5); 
  VSS[0]->SetMinimum(0.0005); 
  VSS[0]->GetXaxis()->SetTitle("N(SameSide)");
  VSS[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("SS Fractions",350,350);
  for (unsigned int iHis=0; iHis < VSS.size(); ++iHis) {
    vroot.addTH1DPtr("SS Fractions",VSS[iHis]);  
  }

  // x4 = Energy fraction of fourth jet (control)
  nBins = 30;
  xLo =  0.0;
  xHi =  0.6;
  double wx4 = nBins/(xHi-xLo);
  vector<TH1D*> Vx4;
  Vx4.push_back(new TH1D("x4","x_{4}",nBins, xLo, xHi));
  Vx4.push_back(new TH1D("x4b","x_{4}",nBins, xLo, xHi));
  Vx4.push_back(new TH1D("x4c","x_{4}",nBins, xLo, xHi));
  Vx4.push_back(new TH1D("x4d","x_{4}",nBins, xLo, xHi));
  Vx4[0]->SetMaximum(6.01); 
  Vx4[0]->SetMinimum(0.0); 
  Vx4[0]->GetXaxis()->SetTitle("x_{4}");
  Vx4[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("x4",350,350);
  for (unsigned int iHis=0; iHis < Vx4.size(); ++iHis) {
    vroot.addTH1DPtr("x4",Vx4[iHis]);  
  }

  // the14 
  nBins =   9;
  xLo   =   0.0;
  xHi   = 180.0;
  double wthe14 = nBins/(xHi-xLo);
  vector<TH1D*> Vthe14;
  Vthe14.push_back(new TH1D("the14","#theta_{14}",nBins, xLo, xHi));
  Vthe14.push_back(new TH1D("the14b","#theta_{14}",nBins, xLo, xHi));
  Vthe14.push_back(new TH1D("the14c","#theta_{14}",nBins, xLo, xHi));
  Vthe14.push_back(new TH1D("the14d","#theta_{14}",nBins, xLo, xHi));
  Vthe14[0]->SetMaximum(0.015); 
  Vthe14[0]->SetMinimum(0.0); 
  Vthe14[0]->GetXaxis()->SetTitle("#theta");
  Vthe14[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("the14",350,350);
  for (unsigned int iHis=0; iHis < Vthe14.size(); ++iHis) {
    vroot.addTH1DPtr("the14",Vthe14[iHis]);  
  }

  // cos14 
  nBins =   20;
  xLo   =  -1.0;
  xHi   =   1.0;
  double wcos14 = nBins/(xHi-xLo);
  vector<TH1D*> Vcos14;
  Vcos14.push_back(new TH1D("cos14","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14.push_back(new TH1D("cos14b","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14.push_back(new TH1D("cos14c","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14.push_back(new TH1D("cos14d","cos(#theta_{14})",nBins, xLo, xHi));
  Vcos14[0]->SetMaximum(1.4); 
  Vcos14[0]->SetMinimum(-0.01); 
  Vcos14[0]->GetXaxis()->SetTitle("cos(#theta)");
  Vcos14[0]->GetYaxis()->SetTitle("Rate");
  // Create VinciaRoot display and add histograms to it
  vroot.createDisplay("cos14",350,350);
  for (unsigned int iHis=0; iHis < Vcos14.size(); ++iHis) {
    vroot.addTH1DPtr("cos14",Vcos14[iHis]);  
  }

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtrA->mode("Random:seed");
  vinciaA.rndmPtr->init(initseed);
  vinciaB.rndmPtr->init(initseed);
  vinciaC.rndmPtr->init(initseed);
  vinciaD.rndmPtr->init(initseed);

  // FastJet needs bool to tell whether this event is the first one
  bool firstEvent = true; 

  // Begin event loop
  double sumNa(0.), sumNb(0.), sumNc(0.), sumNd(0.);
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    if (iEvent%(max(1,nEvent/max(nShow,1))) == 0)  
      cout << " Now begin event " << iEvent 
	   << " <N> = "<<sumNa/max(iEvent,1)<<" "<<sumNb/max(iEvent,1)
	   <<" "<<sumNc/max(iEvent,1)<<" "<<sumNd/max(iEvent,1)<<endl;

    Pythia* pythiaPtr;
    double eCM;

    // Loop over the two different generators
    for (int iGen = 0; iGen <= 3; ++iGen) {

      // Set event-generator pointer
      if (iGen == 0) {
	pythiaPtr = &pythiaA;
	eCM       = eCMa;
      } else if (iGen == 1) {
	pythiaPtr = &pythiaB;
	eCM       = eCMb;
      } else if (iGen == 2) {
	pythiaPtr = &pythiaC;
	eCM       = eCMc;
      } else {
	pythiaPtr = &pythiaD;
	eCM       = eCMd;
      }

      // Shorthand for event record
      Event& event = pythiaPtr->event;
      
      // Generate event
      if (!pythiaPtr->next()) continue;
    
      // List the first few events
      if (iEvent <= 2) event.list();
      
      // Add final-state particles to FastJet input vector    
      fjInputs.resize(0);
      for (int i=1;i<event.size();i++) {
	
	// Final state only
	if (!event[i].isFinal())        continue;

	// Count final-state charged multiplicity (as cross check)
	if (event[i].isCharged()) {
	  if (iGen == 0) sumNa++;
	  else if (iGen == 1) sumNb++;
	  else if (iGen == 2) sumNc++;
	  else if (iGen == 3) sumNd++;
	}

	// Now decide what to pass to FastJet
	// NOTE: neutrinos are *included* in this list, since ALEPH
	// appears to have corrected their data to include them.
	// Store as input to Fastjet
	fjInputs.push_back( fastjet::PseudoJet (event[i].px(),event[i].py(),
						event[i].pz(),event[i].e() ) ); 
      }
    
      // FastJet
      if (fjInputs.size() == 0) {
	cout << "Error: event with no final state particles" << endl;
	continue;
      }

      // Run Fastjet algorithm    
      fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);

      // For the first event, print the FastJet details
      if (firstEvent) {
	cout << "Ran " << jetDef->description() << endl;
	cout << "Strategy adopted by FastJet was "
	     << clustSeq.strategy_string() << endl << endl;      
	firstEvent = false;
      }

      // Extract merging scales
      double d56   = clustSeq.exclusive_dmerge(5);
      double d45   = clustSeq.exclusive_dmerge(4);
      double y45   = clustSeq.exclusive_ymerge(4);
      double lny45 = log(y45);
      double d34   = clustSeq.exclusive_dmerge(3);
      double y34   = clustSeq.exclusive_ymerge(3);
      double lny34 = log(y34);
      // 23
      double y23   = clustSeq.exclusive_ymerge(2);
      double lny23 = log(y23);
      // Check the subjets inside each of the two final jets
      vector<fastjet::PseudoJet> dijet = clustSeq.exclusive_jets(2);
      double mMin = min(dijet[0].m(),dijet[1].m());
      double mMax = max(dijet[0].m(),dijet[1].m());
      
      // Exclusive 4-jet quantities
      bool is4jet = false;
      double the12 = 0.0;
      double the13 = 0.0;
      double the14 = 0.0;
      double the23 = 0.0;
      double e4    = 0.0;
      if (y34 > 0.005) {      
	is4jet = true;
	vector<fastjet::PseudoJet> quadjet = clustSeq.exclusive_jets(4);	
	// Sort by energy
	vector<fastjet::PseudoJet> quadjetSorted = sorted_by_E(quadjet);
	// Invariant mass of two softest jets
	the12 = jetAngle(quadjetSorted[0],quadjetSorted[1]);
	the13 = jetAngle(quadjetSorted[0],quadjetSorted[2]);
	the14 = jetAngle(quadjetSorted[0],quadjetSorted[3]);
	the23 = jetAngle(quadjetSorted[1],quadjetSorted[2]);
	//the24 = jetAngle(quadjetSorted[1],quadjetSorted[3]);
	//the34 = jetAngle(quadjetSorted[2],quadjetSorted[3]);
	//e3    = quadjetSorted[2].e();
	e4    = quadjetSorted[3].e();
      }
      
      // Check if 1->N last clusterings happen on same side
      bool sameSide13 = true;
      bool sameSide14 = true;
      bool sameSide15 = true;
      double d0_12   = clustSeq.exclusive_subdmerge(dijet[0], 1);
      double d1_12   = clustSeq.exclusive_subdmerge(dijet[1], 1);
      // Check if first subclustering in any of the jets equals second global scale
      if (abs(1-d0_12/d34) < 1e-5 || abs(1-d1_12/d34) < 1e-5 ) {
	sameSide13=false;
	sameSide14=false;
	sameSide15=false;
      }
      // Check if first subclustering in any of the jets equals third global scale
      else if (abs(1-d0_12/d45) < 1e-5 || abs(1-d1_12/d45) < 1e-5 ) {
	sameSide14=false;
	sameSide15=false;
      }
      // Check if first subclustering in any of the jets equals fourth global scale
      else if (abs(1-d0_12/d56) < 1e-5 || abs(1-d1_12/d56) < 1e-5 ) {
	sameSide15=false;
      }

      // Fill histograms (include uncertainty variations for B if on)
      // Event weights 
      double evwt = 1.0;

      // Fill Jet resolution histograms with results from FastJet
      if (y23 > 0.01) {
	if (y34 > 0.5 * y23) {
	  VDM[iGen]->Fill(pow2(mMin/mMax),wDM*evwt);
	}
      }
      if (sameSide13 && y34 > 0.005) {
	VLnY24SS[iGen]->Fill(lny34-lny23,wLnY24SS*evwt);
      }
      if (sameSide14 && y45 > 0.005) {
	VLnY25SS[iGen]->Fill(lny45-lny23,wLnY25SS*evwt);
      }
      if (sameSide15 && y45 > 0.005) VSS[iGen]->Fill(4.,evwt);
      else if (sameSide14 && y34 > 0.005) VSS[iGen]->Fill(3.,evwt);
      else if (sameSide13 && y23 > 0.005) VSS[iGen]->Fill(2.,evwt);
      else VSS[iGen]->Fill(1.,evwt);
      
      // Exclusive 4-jet quantities
      if (is4jet) {
	Vx4[iGen]->Fill(2.*e4/eCM,wx4*evwt);
	// Something approaching Simon's color-coherence variable
	// angle of 4th jet with respect to a collinear 3-parton conf
	// Can still be 3-parton soft 
	if (the12 > 2.1 && the13 > 2.1 && the23 < 0.524) {
	  Vthe14[iGen]->Fill(the14/M_PI*180.,wthe14*evwt);
	  Vcos14[iGen]->Fill(cos(the14),wcos14*evwt);
	}
      }
    }

    // Periodically update displays
    if (iEvent%250 == 249|| iEvent == nEvent-1) 
      vroot.updateDisplay();       
    
    // Save plots every N events
    if (iEvent%10000 == 9999 || iEvent == nEvent-1) 
      vroot.saveDisplays("vincia08","eps");
    
    // Save histograms as dat files for later 
    // vroot.saveHistograms();
    
  }

  // 
  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR A"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaA.statistics();
  // Print out Vincia internal histograms
  if (vinciaOnA) {
    if (verbose >= 2) vinciaA.printHistos();
    vinciaA.printInfo();
  }

  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR B"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaB.statistics();
  // Print out Vincia internal histograms
  if (vinciaOnB) {
    if (verbose >= 2) vinciaB.printHistos();
    vinciaB.printInfo();
  }

  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR C"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaC.statistics();
  // Print out Vincia internal histograms
  if (vinciaOnC) {
    if (verbose >= 2) vinciaC.printHistos();
    vinciaC.printInfo();
  }

  cout<<"------------------------------------------------------------------------\n";
  cout<<"END-OF-RUN GENERATOR D"<<endl;
  cout<<"------------------------------------------------------------------------\n";

  pythiaD.statistics();
  // Print out Vincia internal histograms
  if (vinciaOnD) {
    if (verbose >= 2) vinciaD.printHistos();
    vinciaD.printInfo();
  }

  cout << endl << "Enter q to terminate..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
